# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Basic display of optimization history: functions and x."""
from __future__ import division, unicode_literals

import logging
from typing import Iterable, List, Optional, Sequence, Tuple

import matplotlib.gridspec as gridspec
import pylab
from matplotlib.colors import SymLogNorm
from matplotlib.figure import Figure
from matplotlib.ticker import LogFormatter, MaxNLocator
from numpy import abs as np_abs
from numpy import append, arange, argmin, array, atleast_2d, concatenate
from numpy import e as npe
from numpy import hstack, isnan
from numpy import log10 as np_log10
from numpy import logspace
from numpy import max as np_max
from numpy import min as np_min
from numpy import ndarray, ones, ones_like
from numpy import sort as np_sort
from numpy import vstack, where
from numpy.linalg import norm

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.function import MDOFunction
from gemseo.post.core.colormaps import PARULA, RG_SEISMIC
from gemseo.post.core.hessians import SR1Approx
from gemseo.post.opt_post_processor import OptPostProcessor
from gemseo.utils.py23_compat import PY2

LOGGER = logging.getLogger(__name__)


class OptHistoryView(OptPostProcessor):
    """The **OptHistoryView** post processing performs separated plots: the design
    variables history, the objective function history, the history of hessian
    approximation of the objective, the inequality constraint history, the equality
    constraint history, and constraints histories.

    By default, all design variables are considered. A sublist of design variables can
    be passed as options. Minimum and maximum values for the plot can be passed as
    options. The objective function can also be represented in terms of difference
    w.r.t. the initial value It is possible either to save the plot, to show the plot or
    both.
    """

    def __init__(
        self,
        opt_problem,  # type: OptimizationProblem
    ):  # type: (...) -> None
        super(OptHistoryView, self).__init__(opt_problem)
        self.cmap = PARULA  # "viridis"  # "jet"
        self.ineq_cstr_cmap = RG_SEISMIC  # "seismic" "PRGn_r"
        self.eq_cstr_cmap = "seismic"  # "seismic" "PRGn_r"

    def _plot(
        self,
        variables_names=None,  # type: Optional[Sequence[str]]
        obj_min=None,  # type: Optional[float]
        obj_max=None,  # type: Optional[float]
        obj_relative=False,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            variables_names: The names of the variables to display.
                If None, use all design variables.
            obj_max: The maximum value for the objective in the plot.
                If None, use the maximum value of the objective history.
            obj_min: The minimum value for the objective in the plot.
                If None, use the minimum value of the objective history.
            obj_relative: If True, plot the objective value difference
                with the initial value.
        """
        # compute the names of the inequality and equality constraints
        ineq_cstr = self.opt_problem.get_ineq_constraints()
        ineq_cstr_names = [c.name for c in ineq_cstr]
        eq_cstr = self.opt_problem.get_eq_constraints()
        eq_cstr_names = [c.name for c in eq_cstr]

        obj_name = self.opt_problem.get_objective_name()
        obj_history, x_history, n_iter = self._get_history(obj_name, variables_names)

        # design variables
        self._create_variables_plot(x_history, variables_names)

        # objective function
        self._create_obj_plot(
            obj_history,
            n_iter,
            obj_min,
            obj_max,
            obj_relative=obj_relative,
        )

        self._create_x_star_plot(x_history, n_iter)

        # Hessian plot
        plot_hessian = not self.database.is_func_grad_history_empty(obj_name)
        if plot_hessian:
            self._create_hessian_approx_plot(self.database, obj_name)

        # inequality and equality constraints
        self._plot_cstr_history(ineq_cstr_names, MDOFunction.TYPE_INEQ)
        self._plot_cstr_history(eq_cstr_names, MDOFunction.TYPE_EQ)

    def _plot_cstr_history(
        self,
        cstr_names,  # type: Sequence[str]
        cstr_type,  # type: str
    ):  # type: (...) -> None
        """Create the plot for (in)equality constraints.

        Args:
            cstr_names: The names of the constraints.
            cstr_type: The type of the constraints, either 'eq' or 'ineq'.
        """
        if cstr_names is not None:
            _, constraints_history = self._get_constraints(cstr_names)
            self._create_cstr_plot(
                constraints_history,
                cstr_type,
                cstr_names,
            )

    def _get_history(
        self,
        function_name,  # type: str
        variables_names,  # type: Sequence[str]
    ):  # type: (...) -> Tuple[ndarray,ndarray,int]
        """Access the optimization history of a function and the design variables at
        which it was computed.

        Args:
            function_name: The name of the function.
            variables_names: The names of the variables to display.

        Returns:
            The function values,
            the design variables values
            and the number of iterations.
        """
        f_hist, x_hist = self.database.get_func_history(function_name, x_hist=True)
        f_hist = array(f_hist).real
        x_hist = array(x_hist).real

        if variables_names is not None:
            # select only the interesting columns
            block_list = []
            column = 0
            for var in self.opt_problem.design_space.variables_names:
                if var in variables_names:
                    size = self.opt_problem.design_space.variables_sizes[var]
                    block_list.append(x_hist[:, column : column + size])
                    column += size
            # concatenate the blocks
            x_hist = hstack(block_list)
        n_iter = x_hist.shape[0]
        return f_hist, x_hist, n_iter

    def _get_constraints(
        self,
        cstr_names,  # type: Iterable[str]
    ):  # type: (...) -> Tuple[ndarray,List[ndarray]]
        """Extract a history of constraints.

        Args:
            cstr_names: The names of the constraints.

        Returns:
            The bounds of the constraints and history array.
        """
        available_data_names = self.database.get_all_data_names()
        for cstr_name in cstr_names:
            if cstr_name not in available_data_names:
                cstr_names.remove(cstr_name)

        constraints_history = []
        bnd_list = -1e300 * ones(len(cstr_names))

        for cstr_i, cstr_name in enumerate(cstr_names):

            cstr_history = array(self.database.get_func_history(cstr_name)).real

            constraints_history.append(cstr_history)
            bnd_list[cstr_i] = max(
                bnd_list[cstr_i],
                max(abs(np_min(cstr_history)), abs(np_max(cstr_history))),
            )

        return bnd_list, constraints_history

    def _normalize_x_hist(
        self,
        x_history,  # type: ndarray
        variables_names,  # type: Sequence[str]
    ):  # type: (...) -> ndarray
        """Normalize the design variables history.

        Args:
            x_history: The history for the design variables.
            variables_names: The names of the variables to display.

        Returns:
            The normalized design variables array.
        """
        x_hist_n = x_history.copy()
        lower_bounds = self.opt_problem.design_space.get_lower_bounds(variables_names)
        upper_bounds = self.opt_problem.design_space.get_upper_bounds(variables_names)
        norm_coeff = 1 / (np_abs(upper_bounds - lower_bounds))
        for i in range(x_history.shape[0]):
            x_hist_n[i, :] = (x_hist_n[i, :] - lower_bounds) * norm_coeff
        return x_hist_n

    def _create_variables_plot(
        self,
        x_history,  # type: ndarray
        variables_names,  # type: Sequence[str]
    ):  # type: (...) -> None
        """Create the design variables plot.

        Args:
             x_history: The history for the design variables.
             variables_names: The names of the variables to display.
        """
        if len(x_history) < 2:
            return
        n_variables = x_history.shape[1]
        norm_x_history = self._normalize_x_hist(x_history, variables_names)

        fig = pylab.plt.figure(figsize=(11, 6))
        grid = gridspec.GridSpec(1, 2, width_ratios=[15, 1], wspace=0.04, hspace=0.6)

        # design variables
        ax1 = fig.add_subplot(grid[0, 0])
        im1 = ax1.imshow(
            norm_x_history.T,
            cmap=self.cmap,
            interpolation="nearest",
            vmin=0.0,
            vmax=1.0,
            aspect="auto",
        )
        # y ticks: names of the variables
        design_space = self.opt_problem.design_space
        y_labels = []
        if variables_names is None:
            variables_names = design_space.variables_names
        for variable_name in variables_names:
            size = design_space.variables_sizes[variable_name]
            name = variable_name
            if size > 1:
                name += " (0)"
            y_labels.append(name)
            for i in range(1, size):
                y_labels.append("({})".format(i))

        ax1.set_yticks(arange(n_variables))
        ax1.set_yticklabels(y_labels)
        # ax1.invert_yaxis()

        ax1.set_title("Evolution of the optimization variables")
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))

        # colorbar
        ax2 = fig.add_subplot(grid[0, 1])
        fig.colorbar(im1, cax=ax2)

        # Set window size
        mng = pylab.plt.get_current_fig_manager()
        mng.resize(700, 1000)

        self._add_figure(fig, "variables")

    def _create_obj_plot(
        self,
        obj_history,  # type: ndarray
        n_iter,  # type: int
        obj_min=None,  # type: Optional[float]
        obj_max=None,  # type: Optional[float]
        obj_relative=False,  # type: bool
    ):  # type: (...) -> None
        """Creates the design variables plot.

        Args:
            obj_history: The history of the objective function.
            n_iter: The number of iterations.
            obj_max: The maximum value for the objective in the plot.
                If None, use the maximum value of the objective history.
            obj_min: The minimum value for the objective in the plot.
                If None, use the minimum value of the objective history.
            obj_relative: If True, plot the objective value difference
                with the initial value.
        """
        # if max problem, plot -objective value
        if not self.opt_problem.minimize_objective:
            obj_history = -obj_history
        if obj_relative:
            LOGGER.info(
                "Plot of optimization history "
                "with relative variation compared to "
                "initial point objective value = %s",
                obj_history[0],
            )
            obj_history -= obj_history[0]
        # Remove nans
        n_x = len(obj_history)
        x_absc = arange(n_x)
        x_absc_nan = None
        idx_nan = isnan(obj_history)

        if idx_nan.size > 0:
            obj_history = obj_history[~idx_nan]
            x_absc_nan = x_absc[idx_nan]
            x_absc = x_absc[~idx_nan]

        fmin = np_min(obj_history)
        fmax = np_max(obj_history)

        fig = pylab.plt.figure(figsize=(11, 6))
        # objective function
        pylab.plt.xlabel("Iterations", fontsize=12)
        pylab.plt.ylabel("Objective value", fontsize=12)

        pylab.plt.plot(x_absc, obj_history)

        if idx_nan.size > 0:
            for x_i in x_absc_nan:
                pylab.plt.axvline(x_i, color="purple")

        if obj_min is not None and obj_min < fmin:
            fmin = obj_min
        if obj_max is not None and obj_max > fmax:
            fmax = obj_max
        pylab.plt.ylim([fmin, fmax])
        pylab.plt.xlim([0, n_iter])
        ax1 = fig.gca()
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))
        pylab.plt.grid(True)
        pylab.plt.title("Evolution of the objective value")

        # Set window size
        mng = pylab.plt.get_current_fig_manager()
        mng.resize(700, 1000)

        self._add_figure(fig, "objective")

    def _create_x_star_plot(
        self,
        x_history,  # type: ndarray
        n_iter,  # type:int
    ):
        """Create the design variables plot.

        Args:
            x_history: The history of the design variables.
            n_iter: The number of iterations.
        """
        fig = pylab.plt.figure(figsize=(11, 6))
        # objective function
        pylab.plt.xlabel("Iterations", fontsize=12)
        pylab.plt.ylabel("||x-x*||", fontsize=12)

        n_i = x_history.shape[0]
        _, x_opt, _, _, _ = self.opt_problem.get_optimum()

        normalize = self.opt_problem.design_space.normalize_vect

        x_xstar = [
            norm(normalize(x_history[i, :]) - normalize(x_opt)) for i in range(n_i)
        ]
        # Draw a vertical line at the optimum
        ind_opt = argmin(x_xstar)
        pylab.plt.axvline(x=ind_opt, color="r")
        pylab.plt.semilogy(arange(len(x_xstar)), x_xstar)
        # ======================================================================
        # try:
        #     pylab.plt.semilogy(np.arange(len(x_xstar)), x_xstar)
        # except ValueError:
        #     LOGGER.warning("Cannot use log scale for x_star plot since" +
        #                    "all values are not positive !")
        # ======================================================================
        ax1 = fig.gca()
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))
        pylab.plt.grid(True)
        pylab.plt.title("Distance to the optimum")
        pylab.plt.xlim([0, n_iter])

        # Set window size
        mng = pylab.plt.get_current_fig_manager()
        mng.resize(700, 1000)

        self._add_figure(fig, "x_xstar")

    @staticmethod
    def _cstr_number(
        cstr_names,  # type: Sequence[str]
        cstr_history,  # type: ndarray
    ):  # type: (...) -> int
        """Compute the total scalar constraints number.

        Args:
            cstr_names: The names of the constraints.
            cstr_history: The history of the constraints.

        Returns:
            The number of constraints.
        """
        n_cstr = 0
        for cstr_i in range(len(cstr_names)):
            c_hist_loc = atleast_2d(cstr_history[cstr_i]).T
            if c_hist_loc.shape[1] == 1:
                c_hist_loc = c_hist_loc.T
            n_cstr += c_hist_loc.shape[0]
        LOGGER.debug("Total constraints number =%s", n_cstr)
        return n_cstr

    def _create_cstr_plot(
        self,
        cstr_history,  # type: ndarray
        cstr_type,  # type: str
        cstr_names,  # type: Sequence[str]
    ):  # type: (...) -> None
        """Create the constraints plot: 1 line per constraint component.

        Args:
            cstr_history: The history of the constraints.
            cstr_type: The type of the constraints, either 'eq' or 'ineq'.
            cstr_names: The names of the constraints.
        """
        n_cstr = self._cstr_number(cstr_names, cstr_history)
        if n_cstr == 0:
            return

        # matrix of all constraints' values
        cstr_matrix = None
        vmax = 0.0
        cstr_labels = []

        max_iter = 0
        for (_, cstr_history_i) in enumerate(cstr_history):
            history_i = atleast_2d(cstr_history_i).T
            if history_i.shape[1] == 1:
                history_i = history_i.T

            if max_iter < history_i.shape[1]:
                max_iter = history_i.shape[1]

        for (i, cstr_history_i) in enumerate(cstr_history):
            history_i = atleast_2d(cstr_history_i).T
            if history_i.shape[1] == 1:
                history_i = history_i.T

            nb_components = history_i.shape[0]

            if history_i.shape[1] == max_iter:  # TEST
                for component_j in range(nb_components):
                    # compute the label of the constraint
                    if component_j == 0:
                        cstr_name = cstr_names[i]
                        if nb_components >= 2:
                            cstr_name += " (" + str(component_j + 1) + ")"
                        cstr_labels.append(cstr_name)
                    else:
                        cstr_labels.append("(" + str(component_j + 1) + ")")

                    history_i_j = atleast_2d(history_i[component_j, :])

                    # max value
                    notnans = ~isnan(history_i_j)
                    vmax = max(vmax, np_max(np_abs(history_i_j[notnans])))

                    # build the constraint matrix
                    if cstr_matrix is None:
                        cstr_matrix = history_i_j
                    else:
                        cstr_matrix = vstack((cstr_matrix, history_i_j))

        fig = self._build_cstr_fig(cstr_matrix, cstr_type, vmax, n_cstr, cstr_labels)

        self._add_figure(fig, "{}_constraints".format(cstr_type))

    def _build_cstr_fig(
        self,
        cstr_matrix,  # type: ndarray
        cstr_type,  # type: str
        vmax,  # type: float
        n_cstr,  # type: int
        cstr_labels,  # type: Sequence[str]
    ):  # type: (...) -> Figure
        """Build the constraints figure.

        Args:
            cstr_matrix: The matrix of constraints values.
            cstr_type: The type of the constraints, either 'eq' or 'ineq'.
            cstr_labels: The labels for the constraints.
            vmax: The maximum constraint absolute value.
            n_cstr: The number of constraints.
            cstr_labels: The labels of constraints names.

        Returns:
            The constraints figure.
        """
        # cmap of the constraints
        fullname = "equality"
        if cstr_type == MDOFunction.TYPE_EQ:
            cmap = self.eq_cstr_cmap
        else:
            cmap = self.ineq_cstr_cmap
            fullname = "in" + fullname

        idx_nan = isnan(cstr_matrix)
        hasnan = idx_nan.any()
        if hasnan > 0:
            cstr_matrix[idx_nan] = 0.0

        # generation of the image
        fig = pylab.plt.figure(figsize=(11, 6))
        grid = gridspec.GridSpec(1, 2, width_ratios=[15, 1], wspace=0.04, hspace=0.6)
        ax1 = fig.add_subplot(grid[0, 0])
        if PY2:
            im1 = ax1.imshow(
                cstr_matrix,
                cmap=cmap,
                interpolation="nearest",
                aspect="auto",
                norm=SymLogNorm(linthresh=1.0, vmin=-vmax, vmax=vmax),
            )
        else:
            im1 = ax1.imshow(
                cstr_matrix,
                cmap=cmap,
                interpolation="nearest",
                aspect="auto",
                norm=SymLogNorm(linthresh=1.0, vmin=-vmax, vmax=vmax, base=npe),
            )
        if hasnan > 0:
            x_absc_nan = where(idx_nan.any(axis=0))[0]
            for x_i in x_absc_nan:
                pylab.plt.axvline(x_i, color="purple")

        ax1.tick_params(labelsize=9)
        ax1.set_yticks(list(range(n_cstr)))
        ax1.set_yticklabels(cstr_labels)

        ax1.set_xlabel("evaluations")
        ax1.set_title("Evolution of the " + fullname + " constraints")
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))

        ax1.hlines(
            list(range(len(cstr_matrix))),
            [-0.5],
            [len(cstr_matrix[0]) - 0.5],
            alpha=0.1,
            lw=0.5,
        )

        # color map
        cax = fig.add_subplot(grid[0, 1])
        if 0.0 < vmax < 1.0:
            thick_min = int(np_log10(vmax))
        else:
            thick_min = 0
        if vmax > 1.0:
            thick_max = int(np_log10(vmax))
        else:
            thick_max = 0
        thick_num = thick_max - thick_min + 1
        levels_pos = logspace(thick_min, thick_max, num=thick_num, base=10.0)
        if vmax != 0.0:
            levels_pos = np_sort(append(levels_pos, vmax))
        levels_neg = np_sort(-levels_pos)
        levels_neg = append(levels_neg, 0)
        levels = concatenate((levels_neg, levels_pos))
        col_bar = fig.colorbar(im1, cax=cax, ticks=levels)
        col_bar.ax.tick_params(labelsize=9)
        cax.set_xlabel("symlog")

        mng = pylab.plt.get_current_fig_manager()
        # mng.full_screen_toggle()
        mng.resize(700, 1000)
        return fig

    def _create_hessian_approx_plot(
        self,
        history,  # type: ndarray
        obj_name,  # type: str
    ):  # type: (...) -> None
        """Create the plot of the Hessian approximation.

        Args:
            history: The optimization history.
            obj_name: The objective function name.
        """
        try:
            approximator = SR1Approx(history)
            _, diag, _, _ = approximator.build_approximation(
                funcname=obj_name, save_diag=True
            )
            diag = [ones_like(diag[0])] + diag  # Add first iteration blank
            diag = array(diag).T
        except ValueError:
            LOGGER.warning("Failed to create Hessian approximation", exc_info=True)
            return

        # if max problem, plot -Hessian
        if not self.opt_problem.minimize_objective:
            diag = -diag

        fig = pylab.plt.figure(figsize=(11, 6))
        grid = gridspec.GridSpec(1, 2, width_ratios=[15, 1], wspace=0.04, hspace=0.6)
        # matrix
        axe = fig.add_subplot(grid[0, 0])

        axe.set_title("Hessian diagonal approximation")
        axe.set_xlabel("Iterations", fontsize=12)
        axe.set_ylabel("Variable id", fontsize=12)
        axe.xaxis.set_major_locator(MaxNLocator(integer=True))
        vmax = max(abs(np_max(diag)), abs(np_min(diag)))
        linthresh = 10 ** (np_log10(vmax) - 5.0)
        if PY2:
            img = axe.imshow(
                diag.real,
                cmap=self.cmap,
                interpolation="nearest",
                aspect="auto",
                norm=SymLogNorm(linthresh=linthresh, vmin=-vmax, vmax=vmax),
            )
        else:
            img = axe.imshow(
                diag.real,
                cmap=self.cmap,
                interpolation="nearest",
                aspect="auto",
                norm=SymLogNorm(linthresh=linthresh, vmin=-vmax, vmax=vmax, base=npe),
            )
        axe.invert_yaxis()

        # colorbar
        vmax = max(abs(np_max(diag)), abs(np_min(diag)))
        thick_min = int(np_log10(linthresh))
        thick_max = int(np_log10(vmax))
        thick_num = thick_max - thick_min + 1
        levels_pos = logspace(thick_min, thick_max, num=thick_num, base=10.0)
        levels_pos = append(levels_pos, vmax)
        levels_neg = np_sort(-levels_pos)
        levels_neg = append(levels_neg, 0)
        levels = concatenate((levels_neg, levels_pos))

        l_f = LogFormatter(base=10, labelOnlyBase=False)
        cax = fig.add_subplot(grid[0, 1])
        cax.invert_yaxis()
        fig.colorbar(img, cax=cax, ticks=levels, format=l_f)

        mng = pylab.plt.get_current_fig_manager()
        # mng.full_screen_toggle()
        mng.resize(700, 1000)
        self._add_figure(fig, "hessian_approximation")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Coupled derivatives calculations
********************************
"""
from __future__ import division, unicode_literals

from collections import defaultdict

import matplotlib.pyplot as plt
from numpy import atleast_2d, concatenate, empty, ones, zeros
from scipy.sparse import dia_matrix, dok_matrix, vstack
from scipy.sparse.csc import csc_matrix
from scipy.sparse.linalg.dsolve.linsolve import factorized
from scipy.sparse.linalg.interface import LinearOperator

from gemseo.utils.linear_solver import LinearSolver


def none_factory():
    """Returns None...

    To be used for defaultdict
    """


def default_dict_factory():
    """Instantiates a defaultdict(None) object."""
    return defaultdict(none_factory)


class JacobianAssembly(object):
    """Assembly of Jacobians Typically, assemble disciplines's Jacobians into a system
    Jacobian."""

    DIRECT_MODE = "direct"
    ADJOINT_MODE = "adjoint"
    AUTO_MODE = "auto"
    REVERSE_MODE = "reverse"
    AVAILABLE_MODES = (DIRECT_MODE, ADJOINT_MODE, AUTO_MODE, REVERSE_MODE)

    # matrix types
    SPARSE = "sparse"
    LINEAR_OPERATOR = "linear_operator"
    AVAILABLE_MAT_TYPES = [SPARSE, LINEAR_OPERATOR]

    # linear solvers
    AVAILABLE_SOLVERS = LinearSolver.AVAILABLE_SOLVERS

    def __init__(self, coupling_structure):
        """Constructor of the assembly.

        :param coupling_structure: the disciplines that form
           the coupled system
        """
        self.coupling_structure = coupling_structure
        self.sizes = {}
        self.disciplines = {}
        self.coupled_system = CoupledSystem()
        self.n_newton_linear_resolutions = 0
        self.linear_solver = LinearSolver()

    def __check_inputs(self, functions, variables, couplings, matrix_type, use_lu_fact):
        """Checks the inputs before differentiation.

        :param functions: the functions to differentiate
        :param variables: the differentiation variables
        :param couplings: the coupling variables
        :param matrix_type: type of matrix for linearization
        :param use_lu_fact: use LU factorization once for all second members
        """
        unknown_dvars = set(variables)
        unknown_outs = set(functions)
        for discipline in self.coupling_structure.disciplines:
            inputs = set(discipline.get_input_data_names())
            outputs = set(discipline.get_output_data_names())
            unknown_outs = unknown_outs - outputs
            unknown_dvars = unknown_dvars - inputs

        if unknown_dvars:
            raise ValueError(
                "Some of the specified variables are not "
                + "inputs of the disciplines: "
                + str(unknown_dvars)
                + " possible inputs are: "
                + str(
                    [
                        disc.get_input_data_names()
                        for disc in self.coupling_structure.disciplines
                    ]
                )
            )

        if unknown_outs:
            raise ValueError(
                "Some outputs are not computed by the disciplines:"
                + str(unknown_outs)
                + " available outputs are: "
                + str(
                    [
                        disc.get_output_data_names()
                        for disc in self.coupling_structure.disciplines
                    ]
                )
            )

        for coupling in set(couplings) & set(variables):
            raise ValueError(
                "Variable "
                + str(coupling)
                + " is both a coupling and a design variable"
            )

        if matrix_type not in self.AVAILABLE_MAT_TYPES:
            raise ValueError(
                "Unknown matrix type "
                + str(matrix_type)
                + ", available ones are "
                + str(self.AVAILABLE_MAT_TYPES)
            )

        if use_lu_fact and matrix_type == self.LINEAR_OPERATOR:
            raise ValueError(
                "Unsupported LU factorization for "
                + "LinearOperators! Please use Sparse matrices"
                + " instead"
            )

    def compute_sizes(self, functions, variables, couplings):
        """Computes the number of scalar functions, variables and couplings.

        :param functions: the functions to differentiate
        :param variables: the differentiation variables
        :param couplings: the coupling variables
        """
        self.sizes = {}
        self.disciplines = {}
        # search for the functions/variables/couplings in the
        # Jacobians of the disciplines

        # functions
        for function in functions:
            discipline = self.coupling_structure.find_discipline(function)
            self.disciplines[function] = discipline
            # get an arbitrary Jacobian and compute the number of rows
            size = next(iter(discipline.jac[function].values())).shape[0]
            self.sizes[function] = size
        # couplings
        for coupling in couplings:
            discipline = self.coupling_structure.find_discipline(coupling)
            self.disciplines[coupling] = discipline
            # get an arbitrary Jacobian and compute the number of rows
            size = next(iter(discipline.jac[coupling].values())).shape[0]
            self.sizes[coupling] = size
        # variables
        for variable in variables:
            for discipline in self.coupling_structure.disciplines:
                if variable not in self.sizes:
                    for variables_dict in discipline.jac.values():
                        jac = variables_dict.get(variable, None)
                        if jac is not None:
                            self.sizes[variable] = jac.shape[1]
                            self.disciplines[variable] = discipline
                            break
            if variable not in self.sizes:
                raise ValueError(
                    "Failed to determine the size of input variable {}".format(variable)
                )

    @staticmethod
    def _check_mode(mode, n_variables, n_functions):
        """Set the differentiation mode (direct or adjoint)

        :param mode: user given mode
        :param n_variables: number of variables
        :param n_functions: number of functions
        """
        if mode == JacobianAssembly.AUTO_MODE:
            if n_variables <= n_functions:
                mode = JacobianAssembly.DIRECT_MODE
            else:
                mode = JacobianAssembly.ADJOINT_MODE
        return mode

    def compute_dimension(self, names):
        """Compute the total number of functions/variables/couplings of the whole
        system.

        :param names: list of names of inputs or outputs
        """
        number = 0
        for name in names:
            number += self.sizes[name]
        return number

    def _dres_dvar_sparse(self, residuals, variables, n_residuals, n_variables):
        """Forms the matrix of partial derivatives of residuals
        Given disciplinary Jacobians dYi(Y0...Yn)/dvj,
        fill the sparse Jacobian:
        |           |
        |  dRi/dvj  |
        |           |

        :param residuals: the residuals (R)
        :param variables: the differentiation variables
        :param n_residuals: number of residuals
        :param n_variables: number of variables
        """
        dres_dvar = dok_matrix((n_residuals, n_variables))

        out_i = 0
        # Row blocks
        for residual in residuals:
            residual_size = self.sizes[residual]
            # Find the associated discipline
            discipline = self.disciplines[residual]
            residual_jac = discipline.jac[residual]
            # Column blocks
            out_j = 0
            for variable in variables:
                variable_size = self.sizes[variable]
                if residual == variable:
                    # residual Yi-Yi: put -I in the Jacobian
                    ones_mat = (ones(variable_size), 0)
                    shape = (variable_size, variable_size)
                    diag_mat = -dia_matrix(ones_mat, shape=shape)

                    if self.coupling_structure.is_self_coupled(discipline):
                        jac = residual_jac.get(variable, None)
                        if jac is not None:
                            diag_mat += jac
                    dres_dvar[
                        out_i : out_i + variable_size, out_j : out_j + variable_size
                    ] = diag_mat

                else:
                    # block Jacobian
                    jac = residual_jac.get(variable, None)
                    if jac is not None:
                        n_i, n_j = jac.shape
                        assert n_i == residual_size
                        assert n_j == variable_size
                        # Fill the sparse Jacobian block
                        dres_dvar[out_i : out_i + n_i, out_j : out_j + n_j] = jac
                # Shift the column by block width
                out_j += variable_size
            # Shift the row by block height
            out_i += residual_size
        return dres_dvar.real

    def _dres_dvar_linop(self, residuals, variables, n_residuals, n_variables):
        """Forms the linear operator of partial derivatives of residuals.

        :param residuals: the residuals (R)
        :param variables: the differentiation variables
        :param n_residuals: number of residuals
        :param n_variables: number of variables
        """
        # define the linear function
        def dres_dvar(x_array):
            """The linear operator that represents the square matrix dR/dy.

            :param x_array: vector multiplied by the matrix
            """
            assert x_array.shape[0] == n_variables
            # initialize the result
            result = zeros(n_residuals)

            out_i = 0
            # Row blocks
            for residual in residuals:
                residual_size = self.sizes[residual]
                # Find the associated discipline
                discipline = self.disciplines[residual]
                residual_jac = discipline.jac[residual]
                # Column blocks
                out_j = 0
                for variable in variables:
                    variable_size = self.sizes[variable]
                    if residual == variable:
                        # residual Yi-Yi: (-I).x = -x
                        sub_x = x_array[out_j : out_j + variable_size]
                        result[out_i : out_i + residual_size] -= sub_x
                        if self.coupling_structure.is_self_coupled(discipline):
                            jac = residual_jac.get(variable, None)
                            if jac is not None:
                                result[out_i : out_i + residual_size] += jac.dot(sub_x)
                    else:
                        # block Jacobian
                        jac = residual_jac.get(variable, None)
                        if jac is not None:
                            sub_x = x_array[out_j : out_j + variable_size]
                            sub_result = jac.dot(sub_x)
                            result[out_i : out_i + residual_size] += sub_result
                    # Shift the column by block width
                    out_j += variable_size
                # Shift the row by block height
                out_i += residual_size
            return result

        return LinearOperator((n_residuals, n_variables), matvec=dres_dvar)

    def _dres_dvar_t_linop(self, residuals, variables, n_residuals, n_variables):
        """Forms the transposed linear operator of partial derivatives of residuals.

        :param residuals: the residuals (R)
        :param variables: the differentiation variables
        :param n_residuals: number of residuals
        :param n_variables: number of variables
        """
        # define the linear function
        def dres_t_dvar(x_array):
            """The transposed linear operator that represents the square matrix dR/dy.

            :param x_array: vector multiplied by the matrix
            """
            assert x_array.shape[0] == n_residuals
            # initialize the result
            result = zeros(n_variables)

            out_j = 0
            # Column blocks
            for residual in residuals:
                residual_size = self.sizes[residual]
                # Find the associated discipline
                discipline = self.disciplines[residual]
                residual_jac = discipline.jac[residual]
                # Row blocks
                out_i = 0
                for variable in variables:
                    variable_size = self.sizes[variable]
                    if residual == variable:
                        # residual Yi-Yi: (-I).x = -x
                        sub_x = x_array[out_j : out_j + residual_size]
                        result[out_i : out_i + variable_size] -= sub_x
                        if self.coupling_structure.is_self_coupled(discipline):
                            jac = residual_jac.get(variable, None)
                            if jac is not None:
                                result[out_i : out_i + residual_size] += jac.T.dot(
                                    sub_x
                                )
                    else:
                        # block Jacobian
                        jac = residual_jac.get(variable, None)
                        if jac is not None:
                            sub_x = x_array[out_j : out_j + residual_size]
                            sub_result = jac.T.dot(sub_x)
                            result[out_i : out_i + variable_size] += sub_result
                    # Shift the column by block width
                    out_i += variable_size
                # Shift the row by block height
                out_j += residual_size
            return result

        return LinearOperator((n_variables, n_residuals), matvec=dres_t_dvar)

    def dres_dvar(
        self,
        residuals,
        variables,
        n_residuals,
        n_variables,
        matrix_type=SPARSE,
        transpose=False,
    ):
        """Forms the matrix of partial derivatives of residuals
        Given disciplinary Jacobians dYi(Y0...Yn)/dvj,
        fill the sparse Jacobian:
        |           |
        |  dRi/dvj  |
        |           | (Default value = False)

        :param residuals: the residuals (R)
        :param variables: the differentiation variables
        :param n_residuals: number of residuals
        :param n_variables: number of variables
        :param matrix_type: type of the matrix (Default value = SPARSE)
        :param transpose: if True, transpose the matrix
        """
        if matrix_type == JacobianAssembly.SPARSE:
            sparse_dres_dvar = self._dres_dvar_sparse(
                residuals, variables, n_residuals, n_variables
            )
            if transpose:
                return sparse_dres_dvar.T
            return sparse_dres_dvar

        if matrix_type == JacobianAssembly.LINEAR_OPERATOR:
            if transpose:
                return self._dres_dvar_t_linop(
                    residuals, variables, n_residuals, n_variables
                )
            return self._dres_dvar_linop(residuals, variables, n_residuals, n_variables)

        raise TypeError("cannot handle the matrix type")

    def dfun_dvar(self, function, variables, n_variables):
        """Forms the matrix of partial derivatives of a function
        Given disciplinary Jacobians dJi(v0...vn)/dvj,
        fill the sparse Jacobian:
        |           |
        |  dJi/dvj  |
        |           |

        :param function: the function to differentiate
        :param variables: the differentiation variables
        :param n_variables: number of variables
        :returns: the full Jacobian matrix
        """
        function_size = self.sizes[function]
        dfun_dy = dok_matrix((function_size, n_variables))
        # Find the associated discipline
        discipline = self.disciplines[function]
        function_jac = discipline.jac[function]

        # Loop over differentiation variable
        out_j = 0
        for variable in variables:
            variable_size = self.sizes[variable]
            jac_var = function_jac.get(variable, None)
            if jac_var is not None:
                n_i, n_j = jac_var.shape
                assert n_j == variable_size
                assert n_i == function_size
                # Fill the sparse Jacobian block
                dfun_dy[:, out_j : out_j + n_j] = jac_var
            # Shift the column by block width
            out_j += variable_size
        return dfun_dy

    def total_derivatives(
        self,
        in_data,
        functions,
        variables,
        couplings,
        linear_solver="lgmres",
        mode=AUTO_MODE,
        matrix_type=SPARSE,
        use_lu_fact=False,
        exec_cache_tol=None,
        force_no_exec=False,
        **kwargs
    ):
        """Computes the Jacobian of total derivatives of the coupled system formed by
        the disciplines.

        :param in_data: input data dict
        :param functions: the functions to differentiate
        :param variables: the differentiation variables
        :param couplings: the coupling variables
        :param linear_solver: name of the linear solver
            (Default value = 'lgmres')
        :param mode: linearization mode (auto, direct or adjoint)
            (Default value = AUTO_MODE)
        :param matrix_type: representation of the matrix dR/dy (sparse or
            linear operator) (Default value = SPARSE)
        :param use_lu_fact: if True, factorize dres_dy once
            (Default value = False), unsupported for linear operator mode
        :param force_no_exec: if True, the discipline is not
            re executed, cache is loaded anyway
        :param kwargs: dict of optional parameters
        :returns: the dictionary of dictionary of coupled (total) derivatives
        """
        if not functions:
            return defaultdict(default_dict_factory)
        self.__check_inputs(functions, variables, couplings, matrix_type, use_lu_fact)

        # linearize all the disciplines
        self._add_differentiated_inouts(functions, variables, couplings)
        for disc in self.coupling_structure.disciplines:
            if exec_cache_tol is not None:
                disc.cache_tol = exec_cache_tol
            disc.linearize(in_data, force_no_exec=force_no_exec)

        # compute the sizes from the Jacobians
        self.compute_sizes(functions, variables, couplings)
        n_variables = self.compute_dimension(variables)
        n_functions = self.compute_dimension(functions)
        n_couplings = self.compute_dimension(couplings)

        # compute the partial derivatives of the residuals
        dres_dx = self.dres_dvar(couplings, variables, n_couplings, n_variables)

        # compute the partial derivatives of the interest functions
        (dfun_dx, dfun_dy) = ({}, {})
        for fun in functions:
            dfun_dx[fun] = self.dfun_dvar(fun, variables, n_variables)
            dfun_dy[fun] = self.dfun_dvar(fun, couplings, n_couplings)

        mode = self._check_mode(mode, n_variables, n_functions)

        # compute the total derivatives
        if mode == JacobianAssembly.DIRECT_MODE:
            # sparse square matrix dR/dy
            dres_dy = self.dres_dvar(
                couplings, couplings, n_couplings, n_couplings, matrix_type=matrix_type
            )
            # compute the coupled derivatives
            total_derivatives = self.coupled_system.direct_mode(
                functions,
                n_variables,
                n_couplings,
                dres_dx,
                dres_dy,
                dfun_dx,
                dfun_dy,
                linear_solver,
                use_lu_fact=use_lu_fact,
                **kwargs
            )
        elif mode == JacobianAssembly.ADJOINT_MODE:
            # transposed square matrix dR/dy^T
            dres_dy_t = self.dres_dvar(
                couplings,
                couplings,
                n_couplings,
                n_couplings,
                matrix_type=matrix_type,
                transpose=True,
            )
            # compute the coupled derivatives
            total_derivatives = self.coupled_system.adjoint_mode(
                functions,
                dres_dx,
                dres_dy_t,
                dfun_dx,
                dfun_dy,
                linear_solver,
                use_lu_fact=use_lu_fact,
                **kwargs
            )
        else:
            raise ValueError("Incorrect linearization mode " + str(mode))

        return self.split_jac(total_derivatives, variables)

    def _add_differentiated_inouts(self, functions, variables, couplings):
        """Adds functions to the list of differentiated outputs of all disciplines wrt
        couplings, and variables of the discipline.

        :param functions: the functions to differentiate
        :param variables: the differentiation variables
        :param couplings: the coupling variables
        """
        couplings_and_functions = set(couplings) | set(functions)
        couplings_and_variables = set(couplings) | set(variables)

        for discipline in self.coupling_structure.disciplines:
            # outputs
            disc_outputs = discipline.get_output_data_names()
            outputs = list(couplings_and_functions & set(disc_outputs))

            # inputs
            disc_inputs = discipline.get_input_data_names()
            inputs = list(set(disc_inputs) & couplings_and_variables)

            if inputs and outputs:
                discipline.add_differentiated_inputs(inputs)
                discipline.add_differentiated_outputs(outputs)
            if outputs and not inputs:
                base_msg = (
                    "Discipline '{}' has the outputs '{}' that must be "
                    "differenciated, but no coupling or design "
                    "variables as inputs"
                )
                raise ValueError(base_msg.format(discipline.name, outputs))

    def split_jac(self, coupled_system, variables):
        """Splits a Jacobian dict into a dict of dict.

        :param coupled_system: the derivatives to split
        :param variables: variables wrt wich the differentiation is performed
        :returns: the Jacobian as a dict of dict
        """
        j_split = {}
        for function, function_jac in coupled_system.items():
            i_out = 0
            sub_jac = {}
            for variable in variables:
                size = self.sizes[variable]
                sub_jac[variable] = function_jac[:, i_out : i_out + size]
                i_out += size
            j_split[function] = sub_jac
        return j_split

    # Newton step computation
    def compute_newton_step(
        self,
        in_data,
        couplings,
        relax_factor,
        linear_solver="lgmres",
        matrix_type=LINEAR_OPERATOR,
        **kwargs
    ):
        """Compute Newton step for the the coupled system of residuals formed by the
        disciplines.

        :param in_data: input data dict
        :param couplings: the coupling variables
        :param relax_factor: the relaxation factor
        :param linear_solver: the name of the linear solver
            (Default value = 'lgmres')
        :param matrix_type: representation of the matrix dR/dy (sparse or
            linear operator) (Default value = LINEAR_OPERATOR)
        :param kwargs: optional parameters for the linear solver
        :returns: The Newton step -[dR/dy]^-1 . R as a dict of steps
            per coupling variable
        """
        # linearize the disciplines
        self._add_differentiated_inouts(couplings, couplings, couplings)
        for disc in self.coupling_structure.disciplines:
            disc.linearize(in_data)

        self.compute_sizes(couplings, couplings, couplings)
        n_couplings = self.compute_dimension(couplings)

        # compute the partial derivatives of the residuals
        dres_dy = self.dres_dvar(
            couplings, couplings, n_couplings, n_couplings, matrix_type=matrix_type
        )
        # form the residuals
        res = self.residuals(in_data, couplings)
        # solve the linear system
        newton_step = self.linear_solver.solve(
            dres_dy, -relax_factor * res, linear_solver=linear_solver, **kwargs
        )[:, 0]
        self.n_newton_linear_resolutions += 1

        # split the array of steps
        newton_step_dict = {}
        component = 0
        for coupling in couplings:
            size = self.sizes[coupling]
            newton_step_dict[coupling] = newton_step[component : component + size]
            component += size
        return newton_step_dict

    def residuals(self, in_data, var_names):
        """Forms the matrix of residuals wrt coupling variables
        Given disciplinary explicit calculations Yi(Y0_t,...Yn_t),
        fill the residual matrix:

        ::

            [Y0(Y0_t,...Yn_t) - Y0_t]
            [                       ]
            [Yn(Y0_t,...Yn_t) - Yn_t]

        :param in_data: dictionary of values prescribed for the calculation
            of the residuals (Y0_t,...Yn_t)
        :param var_names: names of variables associated with the residuals (R)
        """
        residual_list = []
        # Build rows blocks
        for var in var_names:
            for discipline in self.coupling_structure.disciplines:
                # Find associated discipline
                if var in discipline.get_output_data_names():
                    discipline_output = discipline.get_outputs_by_name(var)
                    residual = atleast_2d(discipline_output - in_data[var])
                    residual_list.append(residual)
        residual_array = concatenate(residual_list, axis=1)[0, :]
        return residual_array

    # plot method
    def plot_dependency_jacobian(
        self,
        functions=None,
        variables=None,
        save=True,
        show=False,
        filepath=None,
        markersize=None,
    ):
        """Plot the Jacobian matrix Nonzero elements of the sparse matrix are
        represented by blue squares.

        :param functions: list of variables (Default value = None)
        :param variables: list of variables (Default value = None)
        :param show: if True, the plot is displayed (Default value = False)
        :param save: if True, the plot is saved in a PDF file (Default
            value = True)
        :param filepath: file path of the saved PDF
        """
        self.compute_sizes(functions, variables, [])
        n_variables = self.compute_dimension(variables)

        total_jac = None
        # compute the positions of the outputs
        outputs_positions = {}
        current_position = 0

        for fun in functions:
            dfun_dx = self.dfun_dvar(fun, variables, n_variables)
            outputs_positions[fun] = current_position
            current_position += self.sizes[fun]

            if total_jac is None:
                total_jac = dfun_dx
            else:
                total_jac = vstack((total_jac, dfun_dx))

        # compute the positions of the inputs
        inputs_positions = {}
        current_position = 0
        for variable in variables:
            inputs_positions[variable] = current_position
            current_position += self.sizes[variable]

        # plot the (sparse) matrix

        fig = plt.figure(figsize=(6, 10))
        ax1 = fig.add_subplot(111)
        plt.spy(total_jac, markersize=markersize)
        ax1.set_aspect("auto")

        plt.yticks(list(outputs_positions.values()), list(outputs_positions.keys()))
        plt.xticks(
            list(inputs_positions.values()), list(inputs_positions.keys()), rotation=90
        )

        filename = None
        if save:
            if filepath is None:
                filename = "coupled_jacobian.pdf"
            else:
                filename = "coupled_jacobian_" + filepath + ".pdf"
            plt.savefig(filename)

        if show:
            plt.show()
        else:
            plt.close()

        return filename


class CoupledSystem(object):
    """This class contains methods that compute coupled (total) derivatives of a system
    of residuals, using several methods:

    - direct or adjoint
    - factorized for multiple RHS
    """

    def __init__(self):
        """Constructor."""
        self.n_linear_resolutions = 0
        self.n_direct_modes = 0
        self.n_adjoint_modes = 0
        self.lu_fact = 0
        self.linear_solver = LinearSolver()

    def direct_mode(
        self,
        functions,
        n_variables,
        n_couplings,
        dres_dx,
        dres_dy,
        dfun_dx,
        dfun_dy,
        linear_solver,
        use_lu_fact,
        **kwargs
    ):
        """Computation of total derivative Jacobian in direct mode.

        :param functions: functions to differentiate
        :param n_variables: number of variables
        :param n_couplings: number of couplings
        :param dres_dx: Jacobian of residuals wrt design variables
        :param dres_dy: Jacobian of residuals wrt coupling variables
        :param dfun_dx: Jacobian of functions wrt design variables
        :param dfun_dy: Jacobian of functions wrt coupling variables
        :param linear_solver: name of the linear solver
        :param use_lu_fact: if True, factorize dres_dy once
        :param kwargs: optional parameters
        :type kwargs: dict
        """
        self.n_direct_modes += 1
        if use_lu_fact:
            return self._direct_mode_lu(
                functions, n_variables, n_couplings, dres_dx, dres_dy, dfun_dx, dfun_dy
            )
        return self._direct_mode(
            functions,
            n_variables,
            n_couplings,
            dres_dx,
            dres_dy,
            dfun_dx,
            dfun_dy,
            linear_solver,
            **kwargs
        )

    def adjoint_mode(
        self,
        functions,
        dres_dx,
        dres_dy_t,
        dfun_dx,
        dfun_dy,
        linear_solver,
        use_lu_fact,
        **kwargs
    ):
        """Computation of total derivative Jacobian in adjoint mode.

        :param functions: functions to differentiate
        :param dres_dx: Jacobian of residuals wrt design variables
        :param dres_dy: Jacobian of residuals wrt coupling variables
        :param dfun_dx: Jacobian of functions wrt design variables
        :param dfun_dy: Jacobian of functions wrt coupling variables
        :param linear_solver: name of the linear solver
        :param use_lu_fact: if True, factorize dres_dy_t once
        :param kwargs: optional parameters
        :type kwargs: dict
        :param dres_dy_t: param kwargs
        """
        self.n_adjoint_modes += 1
        if use_lu_fact:
            return self._adjoint_mode_lu(
                functions, dres_dx, dres_dy_t, dfun_dx, dfun_dy
            )
        return self._adjoint_mode(
            functions, dres_dx, dres_dy_t, dfun_dx, dfun_dy, linear_solver, **kwargs
        )

    def _direct_mode(
        self,
        functions,
        n_variables,
        n_couplings,
        dres_dx,
        dres_dy,
        dfun_dx,
        dfun_dy,
        linear_solver,
        **kwargs
    ):
        """Computation of total derivative Jacobian in direct mode.

        :param functions: functions to differentiate
        :param n_variables: number of variables
        :param n_couplings: number of couplings
        :param dres_dx: Jacobian of residuals wrt design variables
        :param dres_dy: Jacobian of residuals wrt coupling variables
        :param dfun_dx: Jacobian of functions wrt design variables
        :param dfun_dy: Jacobian of functions wrt coupling variables
        :param linear_solver: name of the linear solver
        :param kwargs: optional parameters
        :type kwargs: dict
        """
        # compute the total derivative dy/dx, independent of the
        # function to differentiate
        dy_dx = empty((n_couplings, n_variables))
        self.linear_solver.outer_v = []
        for var_index in range(n_variables):
            dy_dx[:, var_index] = self.linear_solver.solve(
                dres_dy, -dres_dx[:, var_index], linear_solver=linear_solver, **kwargs
            )[:, 0]
            self.n_linear_resolutions += 1
        # assemble the total derivatives of the functions using dy_dx
        jac = {}
        for fun in functions:
            jac[fun] = dfun_dx[fun].toarray() + dfun_dy[fun].dot(dy_dx)
        return jac

    def _adjoint_mode(
        self, functions, dres_dx, dres_dy_t, dfun_dx, dfun_dy, linear_solver, **kwargs
    ):
        """Computation of total derivative Jacobian in adjoint mode.

        :param functions: functions to differentiate
        :param dres_dx: Jacobian of residuals wrt design variables
        :param dres_dy: Jacobian of residuals wrt coupling variables
        :param dfun_dx: Jacobian of functions wrt design variables
        :param dfun_dy: Jacobian of functions wrt coupling variables
        :param linear_solver: name of the linear solver
        :param kwargs: optional parameters
        :type kwargs: dict
        :param dres_dy_t: derivatives of the residuals wrt coupling vars
        """
        jac = {}
        # adjoint vector for each interest function
        self.linear_solver.outer_v = []
        for fun in functions:
            dfunction_dx = dfun_dx[fun]
            dfunction_dy = dfun_dy[fun]
            jac[fun] = empty(dfunction_dx.shape)
            # compute adjoint vector for each component of the function
            for fun_component in range(dfunction_dy.shape[0]):
                adjoint = self.linear_solver.solve(
                    dres_dy_t,
                    -dfunction_dy[fun_component, :].T,
                    linear_solver=linear_solver,
                    **kwargs
                )
                self.n_linear_resolutions += 1
                jac[fun][fun_component, :] = (
                    dfunction_dx[fun_component, :] + (dres_dx.T.dot(adjoint)).T
                )
        return jac

    def _direct_mode_lu(
        self, functions, n_variables, n_couplings, dres_dx, dres_dy, dfun_dx, dfun_dy
    ):
        """Computation of total derivative Jacobian in direct mode.

        :param functions: functions to differentiate
        :param n_variables: number of variables
        :param n_couplings: number of couplings
        :param dres_dx: Jacobian of residuals wrt design variables
        :param dres_dy: Jacobian of residuals wrt coupling variables
        :param dfun_dx: Jacobian of functions wrt design variables
        :param dfun_dy: Jacobian of functions wrt coupling variables
        """
        # compute the total derivative dy/dx, independent of the
        # function to differentiate
        dy_dx = empty((n_couplings, n_variables))
        # compute LU decomposition
        solve = factorized(csc_matrix(dres_dy))
        self.lu_fact += 1

        for var_index in range(n_variables):
            rhs = -dres_dx[:, var_index].todense()
            dy_dx[:, var_index] = solve(rhs).squeeze()
            self.n_linear_resolutions += 1
        # assemble the total derivatives of the functions using dy_dx
        jac = {}
        for fun in functions:
            jac[fun] = dfun_dx[fun].toarray() + dfun_dy[fun].dot(dy_dx)
        return jac

    def _adjoint_mode_lu(self, functions, dres_dx, dres_dy_t, dfun_dx, dfun_dy):
        """Computation of total derivative Jacobian in adjoint mode.

        :param functions: functions to differentiate
        :param dres_dx: Jacobian of residuals wrt design variables
        :param dfun_dx: Jacobian of functions wrt design variables
        :param dfun_dy: Jacobian of functions wrt coupling variables
        :param dres_dy_t:
        """
        jac = {}
        # compute LU factorization
        solve = factorized(dres_dy_t)
        self.lu_fact += 1
        # adjoint vector for each interest function
        for fun in functions:
            dfunction_dx = dfun_dx[fun]
            dfunction_dy = dfun_dy[fun]
            jac[fun] = empty(dfunction_dx.shape)
            # compute adjoint vector for each component of the function
            for fun_component in range(dfunction_dy.shape[0]):
                adjoint = solve(-dfunction_dy[fun_component, :].todense().T)
                self.n_linear_resolutions += 1
                jac[fun][fun_component, :] = (
                    dfunction_dx[fun_component, :] + (dres_dx.T.dot(adjoint)).T
                )
        return jac



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import division, unicode_literals

import logging

from gemseo.utils.string_tools import MultiLineString

"""
Optimization result
*******************
"""


LOGGER = logging.getLogger(__name__)


class OptimizationResult(object):

    """Stores optimization results."""

    DICT_REPR_ATTR = [
        "x_0",
        "x_opt",
        "optimizer_name",
        "message",
        "f_opt",
        "status",
        "n_obj_call",
        "n_grad_call",
        "n_constr_call",
        "is_feasible",
    ]

    HDF_CSTR_GRAD_KEY = "constr_grad:"
    HDF_CSTR_KEY = "constr:"

    def __init__(
        self,
        x_0=None,
        x_opt=None,
        f_opt=None,
        status=None,
        constraints_values=None,
        constraints_grad=None,
        optimizer_name=None,
        message=None,
        n_obj_call=None,
        n_grad_call=None,
        n_constr_call=None,
        is_feasible=False,
    ):
        """Initialize optimization results.

        :param x_0: initial guess for design variables
        :param x_opt: optimal design variables values
        :param f_opt: the objective function values at optimum
        :param status: the optimizer status
        :param message: the optimizer message
        :param n_obj_call: number of call to objective function by optimizer
        :param n_grad_call: number of call to gradient function by optimizer
        :param n_constr_call: number of call to constraints
                function by optimizer
        :param is_feasible: True if the solution is feasible, false else
        """
        self.x_0 = x_0
        self.optimizer_name = optimizer_name
        self.x_opt = x_opt
        self.message = message
        self.f_opt = f_opt
        self.constraints_values = constraints_values
        self.constraints_grad = constraints_grad
        self.status = status
        self.n_obj_call = n_obj_call
        self.n_grad_call = n_grad_call
        self.n_constr_call = n_constr_call
        self.is_feasible = is_feasible

    def __repr__(self):
        msg = MultiLineString()
        msg.add("Optimization result:")
        msg.indent()
        msg.add("Design variables: {}", self.x_opt)
        msg.add("Objective function: {}", self.f_opt)
        msg.add("Feasible solution: {}", self.is_feasible)
        return str(msg)

    def __str__(self):
        msg = MultiLineString()
        msg.add("Optimization result:")
        msg.add("Objective value = {}", self.f_opt)
        if self.is_feasible:
            msg.add("The result is feasible.")
        else:
            msg.add("The result is not feasible.")
        msg.add("Status: {}", self.status)
        msg.add("Optimizer message: {}", self.message)
        if self.n_obj_call is not None:
            msg.add(
                "Number of calls to the objective function by the optimizer: {}",
                self.n_obj_call,
            )
        if self.constraints_values and len(self.constraints_values) < 20:
            msg.add("Constraints values w.r.t. 0: ")
            msg.indent()
            for c_name in sorted(self.constraints_values.keys()):
                if c_name.startswith("-"):
                    name = str(c_name[1:])
                    values = str(-self.constraints_values[c_name])
                else:
                    name = str(c_name)
                    values = str(self.constraints_values[c_name])
                msg.add("{} = {}", name, values)
        return str(msg)

    def get_data_dict_repr(self):
        """Returns a dict representation of self for serialization functions are
        removed.

        :returns: a dict with attributes names as keys
        """
        repr_dict = {}
        for attr_name in self.DICT_REPR_ATTR:
            attr = getattr(self, attr_name)
            if attr is not None:
                repr_dict[attr_name] = attr

        cgrad_key = OptimizationResult.HDF_CSTR_GRAD_KEY
        c_key = OptimizationResult.HDF_CSTR_KEY
        for cstr, cval in self.constraints_values.items():
            repr_dict[c_key + cstr] = cval
        for cstr, cgrad in self.constraints_grad.items():
            repr_dict[cgrad_key + cstr] = cgrad

        return repr_dict

    @staticmethod
    def init_from_dict_repr(**kwargs):
        """Initializes a new opt result from a data dict typically used for
        deserialization.

        :param kwargs: key value pairs from DICT_REPR_ATTR
        """
        allowed = OptimizationResult.DICT_REPR_ATTR
        filt_args = {k: v for k, v in kwargs.items() if k in allowed}
        non_allowed = {k: v for k, v in kwargs.items() if k not in filt_args}
        constraints_values = {}
        constraints_grad = {}
        cgrad_key = OptimizationResult.HDF_CSTR_GRAD_KEY
        c_key = OptimizationResult.HDF_CSTR_KEY
        for attr, val in non_allowed.items():
            if attr.startswith(cgrad_key):
                key_clean = attr.replace(cgrad_key, "")
                constraints_grad[key_clean] = val
            elif attr.startswith(c_key):
                key_clean = attr.replace(c_key, "")
                constraints_values[key_clean] = val
            else:
                raise ValueError("Unknown attribute : " + str(attr))
        opt_res = OptimizationResult(
            constraints_values=constraints_values,
            constraints_grad=constraints_grad,
            **filt_args
        )
        return opt_res



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Abstract class for the computation and analysis of sensitivity indices.

The purpose of a sensitivity analysis is to
qualify or quantify how the model's uncertain inputs impact its output.

This analysis relies on :class:`.SensitivityAnalysis`
computed from a :class:`.MDODiscipline` representing the model,
a :class:`.ParameterSpace` describing the uncertain parameters
and options associated with a particular concrete class
inheriting from :class:`.SensitivityAnalysis` which is an abstract one.
"""

from __future__ import division, unicode_literals

import logging
from typing import (
    Dict,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import six
from custom_inherit import DocInheritMeta
from matplotlib.figure import Figure
from numpy import array, linspace, ndarray, vstack

from gemseo.algos.doe.doe_lib import DOELibraryOptionType
from gemseo.algos.parameter_space import ParameterSpace
from gemseo.core.dataset import Dataset
from gemseo.core.discipline import MDODiscipline
from gemseo.core.doe_scenario import DOEScenario
from gemseo.post.dataset.bars import BarPlot
from gemseo.post.dataset.curves import Curves
from gemseo.post.dataset.radar_chart import RadarChart
from gemseo.post.dataset.surfaces import Surfaces
from gemseo.utils.file_path_manager import FilePathManager, FileType
from gemseo.utils.matplotlib_figure import save_show_figure
from gemseo.utils.py23_compat import Path

LOGGER = logging.getLogger(__name__)

OutputsType = Union[str, Tuple[str, int], Sequence[Union[str, Tuple[str, int]]]]
IndicesType = Dict[str, Dict[str, ndarray]]


@six.add_metaclass(
    DocInheritMeta(
        abstract_base_class=True,
        style="google_with_merge",
        include_special_methods=True,
    )
)
class SensitivityAnalysis(object):
    """Sensitivity analysis.

    The :class:`.SensitivityAnalysis` class provides both
    the values of :attr:`.SensitivityAnalysis.indices`
    and their graphical representations, from either
    the :meth:`.SensitivityAnalysis.plot` method,
    the :meth:`.SensitivityAnalysis.plot_radar` method or
    the :meth:`.SensitivityAnalysis.plot_bar` method.

    It is also possible to use :meth:`.SensitivityAnalysis.sort_parameters`
    to get the parameters sorted according to :attr:`.SensitivityAnalysis.main_method`.
    The :attr:`.SensitivityAnalysis.main_indices` are indices computed with the latter.

    Lastly, the :meth:`.SensitivityAnalysis.plot_comparison` method allows
    to compare the current :class:`.SensitivityAnalysis` with another one.

    Attributes:
        default_output (list(str)): The default outputs of interest.
        dataset (Dataset): The dataset containing the discipline evaluations.
    """

    DEFAULT_DRIVER = None

    def __init__(
        self,
        discipline,  # type: MDODiscipline
        parameter_space,  # type: ParameterSpace
        n_samples,  # type: int
        algo=None,  # type: Optional[str]
        algo_options=None,  # type: Optional[Mapping]
    ):  # type: (...) -> None  # noqa: D205,D212,D415
        """
        Args:
            discipline: A discipline.
            parameter_space: A parameter space.
            n_samples: A number of samples.
            algo: The name of the DOE algorithm.
                If None, use the :attr:`DEFAULT_DRIVER`.
            algo_options: The options of the DOE algorithm.
        """
        self._algo_name = algo or self.DEFAULT_DRIVER
        self.default_output = list(discipline.get_output_data_names())
        algo_options = algo_options or {}
        self.dataset = self.__sample_discipline(
            discipline, parameter_space, n_samples, **algo_options
        )
        self._main_method = self.__class__.__name__
        default_name = FilePathManager.to_snake_case(self.__class__.__name__)
        self._file_path_manager = FilePathManager(
            FileType.FIGURE, default_name=default_name
        )

    def __sample_discipline(
        self,
        discipline,  # type: MDODiscipline
        parameter_space,  # type: ParameterSpace,
        n_samples,  # type: int
        **options  # type: DOELibraryOptionType
    ):  # type: (...) -> Dataset
        """Sample the discipline and return the dataset.

        Args:
            discipline: A discipline.
            parameter_space: A parameter space.
            n_samples: A number of samples.
            **options: The options for the DOE algorithm.

        Returns:
            The evaluations of the discipline.
        """
        discipline.set_cache_policy(discipline.MEMORY_FULL_CACHE)
        scenario = DOEScenario(
            [discipline], "DisciplinaryOpt", self.default_output[0], parameter_space
        )
        scenario.execute(
            {
                "algo": self._algo_name,
                "n_samples": n_samples,
                "algo_options": options,
            }
        )
        inputs = parameter_space.variables_names
        return discipline.cache.export_to_dataset(inputs_names=inputs)

    @property
    def inputs_names(self):  # type: (...) -> List[str]
        """The name of the inputs."""
        return self.dataset.get_names(self.dataset.INPUT_GROUP)

    def compute_indices(
        self, outputs=None  # type: Optional[Sequence[str]]
    ):  # type: (...) -> Dict[str,IndicesType]
        """Compute the sensitivity indices.

        Args:
            outputs: The outputs
                for which to display sensitivity indices.
                If None, use the default outputs, that are all the discipline outputs.

        Returns:
            The sensitivity indices.

            With the following structure:

            .. code-block:: python

                {
                    "method_name": {
                        "output_name": [
                            {
                                "input_name": data_array,
                            }
                        ]
                    }
                }
        """
        raise NotImplementedError

    @property
    def indices(self):  # type: (...) -> Dict[str,IndicesType]
        """The sensitivity indices.

        With the following structure:

        .. code-block:: python

            {
                "method_name": {
                    "output_name": [
                        {
                            "input_name": data_array,
                        }
                    ]
                }
            }
        """
        raise NotImplementedError

    @property
    def main_method(self):  # type: (...) -> str
        """The name of the main method."""
        return self._main_method

    @main_method.setter
    def main_method(
        self,
        name,  # type: str
    ):  # type: (...) -> NoReturn
        raise NotImplementedError("You cannot change the main method.")

    @property
    def main_indices(self):  # type: (...) -> IndicesType
        """The main sensitivity indices.

        With the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        raise NotImplementedError

    def _outputs_to_tuples(
        self,
        outputs,  # type: OutputsType
    ):  # type: (...) -> List[Tuple[str,int]]
        """Convert the outputs to a list of tuple(str,int).

        Args:
            outputs: The outputs
                for which to display sensitivity indices,
                either a name,
                a list of names,
                a (name, component) tuple,
                a list of such tuples or
                a list mixing such tuples and names.
                When a name is specified, all its components are considered.
                If None, use the default outputs.

        Returns:
            The outputs.

            The outputs are formatted as tuples of the form (name, component),
            where name is the output name and component is the output component.
        """
        if not isinstance(outputs, list):
            outputs = [outputs]

        def get_all(output):
            return [(output, index) for index in range(len(self.main_indices[output]))]

        result = [
            output if isinstance(output, tuple) else get_all(output)
            for output in outputs
        ]
        return [item for sublist in result for item in sublist]

    def sort_parameters(
        self, output  # type: Tuple[str,int]
    ):  # type: (...) -> List[str]
        """Return the parameters sorted in descending order.

        Args:
            output: An output of the form (name, component),
                where name is the output name and component is the output component.

        Returns:
            The input parameters sorted in descending order.
        """
        if not isinstance(output, tuple):
            output = (output, 0)
        output_name, output_component = output
        indices = self.main_indices[output_name][output_component]
        names = [
            name
            for name, _ in sorted(
                list(indices.items()), key=lambda item: item[1].sum(), reverse=True
            )
        ]
        return names

    def plot(
        self,
        output,  # type: Union[str,Tuple[str,int]]
        inputs=None,  # type: Optional[Iterable[str]]
        title=None,  # type: Optional[str]
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        file_format=None,  # type: Optional[str]
    ):  # type: (...) -> None
        """Plot the sensitivity indices.

        Args:
            output: The output
                for which to display sensitivity indices,
                either a name or a tuple of the form (name, component).
                If name, its first component is considered.
            inputs: The inputs to display. If None, display all.
            title: The title of the plot. If None, no title.
            save: If True, save the figure.
            show: If True, show the figure.
            file_path: A file path.
                Either a complete file path, a directory name or a file name.
                If None, use a default file name and a default directory.
                The file extension is inferred from filepath extension, if any.
            file_format: A file format, e.g. 'png', 'pdf', 'svg', ...
                Used when ``file_path`` does not have any extension.
                If None, use a default file extension.
        """
        raise NotImplementedError

    def plot_field(
        self,
        output,  # type: Union[str,Tuple[str,int]]
        mesh=None,  # type: Optional[ndarray]
        inputs=None,  # type: Optional[Iterable[str]]
        title=None,  # type: Optional[str]
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_format=None,  # type: Optional[str]
        properties=None,  # type: Mapping
    ):  # type: (...) -> Union[Curves,Surfaces]
        """Plot the sensitivity indices related to a 1D or 2D functional output.

        The output is considered as a 1D or 2D functional variable,
        according to the shape of the mesh on which it is represented.

        Args:
            output: The output
                for which to display sensitivity indices,
                either a name or a tuple of the form (name, component).
                If name, its first component is considered.
            mesh: The mesh on which the p-length output
                is represented. Either a (1, p) array for a 1D functional output
                or a (2, p) array for a 2D one. If None, assume a 1D functional output.
            inputs: The inputs to display. If None, display all inputs.
            title: The title of the plot. If None, no title is displayed.
            save: If True, save the figure.
            show: If True, show the figure.
            file_path: The path of the file to save the figures.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_extension``.
            directory_path: The path of the directory to save the figures.
                If None, use the current working directory.
            file_name: The name of the file to save the figures.
                If None, use a default one generated by the post-processing.
            file_format: A file extension, e.g. 'png', 'pdf', 'svg', ...
                If None, use a default file extension.
            properties: The general properties of a :class:`.DatasetPlot`.
        """
        dataset = Dataset()
        inputs_names = self._sort_and_filter_input_parameters((output, 0), inputs)
        data = []
        for name in inputs_names:
            data.append([value[name] for value in self.main_indices[output[0]]])
        data = array(data)[:, :, 0]
        dataset.set_from_array(data, [output], sizes={output: data.shape[1]})
        dataset.row_names = inputs_names
        mesh = linspace(0, 1, data.shape[1]) if mesh is None else mesh
        dataset.set_metadata("mesh", mesh)
        mesh_dimension = len(dataset.metadata["mesh"].shape)
        if mesh_dimension == 1:
            plot = Curves(dataset)
            plot.title = title
        elif mesh_dimension == 2:
            plot = Surfaces(dataset)
        else:
            raise NotImplementedError
        plot.execute(
            save=save,
            show=show,
            file_path=file_path,
            file_name=file_name,
            file_format=file_format,
            directory_path=directory_path,
            mesh="mesh",
            variable=output,
            properties=properties,
        )
        return plot

    def plot_bar(
        self,
        outputs,  # type: OutputsType
        inputs=None,  # type: Optional[Iterable[str]]
        title=None,  # type: Optional[str]
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_format=None,  # type: Optional[str]
        **options  # type:int
    ):  # type: (...) -> BarPlot
        """Plot the sensitivity indices on a bar chart.

        This method may consider one or more outputs,
        as well as all inputs (default behavior) or a subset.

        Args:
            outputs: The outputs
                for which to display sensitivity indices,
                either a name,
                a list of names,
                a (name, component) tuple,
                a list of such tuples or
                a list mixing such tuples and names.
                When a name is specified, all its components are considered.
                If None, use the default outputs.
            inputs: The inputs to display. If None, display all.
            title: The title of the plot. If None, no title.
            save: If True, save the figure.
            show: If True, show the figure.
            file_path: The path of the file to save the figures.
                If the extension is missing, use ``file_extension``.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_extension``.
            directory_path: The path of the directory to save the figures.
                If None, use the current working directory.
            file_name: The name of the file to save the figures.
                If None, use a default one generated by the post-processing.
            file_format: A file extension, e.g. 'png', 'pdf', 'svg', ...
                If None, use a default file extension.

        Returns:
            A bar chart representing the sensitivity indices.
        """
        outputs = self._outputs_to_tuples(outputs)
        dataset = Dataset()
        inputs_names = self._sort_and_filter_input_parameters(outputs[0], inputs)
        data = {name: [] for name in inputs_names}
        for output in outputs:
            for name in inputs_names:
                data[name].append(self.main_indices[output[0]][output[1]][name])
        for name in inputs_names:
            dataset.add_variable(name, vstack(data[name]))
        dataset.row_names = [
            "{}({})".format(output[0], output[1]) for output in outputs
        ]
        plot = BarPlot(dataset)
        plot.title = title
        plot.execute(
            save=save,
            show=show,
            file_path=file_path,
            file_name=file_name,
            file_format=file_format,
            directory_path=directory_path,
            **options
        )
        return plot

    def plot_radar(
        self,
        outputs,  # type: OutputsType
        inputs=None,  # type: Optional[Iterable[str]]
        title=None,  # type: Optional[str]
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_format=None,  # type: Optional[str]
        min_radius=None,  # type: Optional[float]
        max_radius=None,  # type: Optional[float]
        **options  # type:bool
    ):  # type: (...) -> RadarChart
        """Plot the sensitivity indices on a radar chart.

        This method may consider one or more outputs,
        as well as all inputs (default behavior) or a subset.

        For visualization purposes,
        it is also possible to change the minimum and maximum radius values.

        Args:
            outputs: The outputs
                for which to display sensitivity indices,
                either a name,
                a list of names,
                a (name, component) tuple,
                a list of such tuples or
                a list mixing such tuples and names.
                When a name is specified, all its components are considered.
                If None, use the default outputs.
            inputs: The inputs to display.
                If None, display all.
            title: The title of the plot. If None, no title.
            save: If True, save the figure.
            show: If True, show the figure.
            file_path: The path of the file to save the figures.
                If the extension is missing, use ``file_extension``.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_extension``.
            directory_path: The path of the directory to save the figures.
                If None, use the current working directory.
            file_name: The name of the file to save the figures.
                If None, use a default one generated by the post-processing.
            file_format: A file extension, e.g. 'png', 'pdf', 'svg', ...
                If None, use a default file extension.
            min_radius: The minimal radial value. If None, from data.
            max_radius: The maximal radial value. If None, from data.

        Returns:
            A radar chart representing the sensitivity indices.
        """
        outputs = self._outputs_to_tuples(outputs)
        dataset = Dataset()
        inputs_names = self._sort_and_filter_input_parameters(outputs[0], inputs)
        data = {name: [] for name in inputs_names}
        for output in outputs:
            for name, value in self.main_indices[output[0]][output[1]].items():
                if name in inputs_names:
                    data[name].append(value)
        for name in inputs_names:
            dataset.add_variable(name, vstack(data[name]))
        dataset.row_names = [
            "{}({})".format(output[0], output[1]) for output in outputs
        ]
        plot = RadarChart(dataset)
        plot.title = title
        plot.rmin = min_radius
        plot.rmax = max_radius
        plot.execute(
            save=save,
            show=show,
            file_path=file_path,
            file_name=file_name,
            file_format=file_format,
            directory_path=directory_path,
            **options
        )
        return plot

    @staticmethod
    def _filter_names(
        names,  # type:Iterable[str],
        names_to_keep,  # type:Iterable[str]
    ):  # type: (...) -> List[str]
        """Sort and filter the names.

        Args:
            names: The original names.
            names_to_keep: The names to keep. If None, keep all.

        Returns:
            The filtered names.
        """
        if names_to_keep is not None:
            names = [item for item in names if item in set(names_to_keep)]
        return names

    def _sort_and_filter_input_parameters(
        self,
        output,  # type: Tuple[str,int]
        inputs_to_keep,  # type: Iterable[str]
    ):  # type: (...) -> List[str]
        """Sort and filter the input parameters.

        Args:
            output: An output for which to display sensitivity indices.
            inputs_to_keep: The inputs to keep. If None, keep all.

        Returns:
            The filtered names.
        """
        inputs = self.sort_parameters((output[0], output[1]))
        return self._filter_names(inputs, inputs_to_keep)

    def plot_comparison(
        self,
        indices,  # type: List[SensitivityAnalysis]
        output,  # type: Union[str,Tuple[str,int]]
        inputs=None,  # type: Optional[Iterable[str]]
        title=None,  # type: Optional[str]
        use_bar_plot=True,  # type: bool
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_format=None,  # type: Optional[str]
        **options  # type:bool
    ):  # type: (...) -> Union[BarPlot,RadarChart]
        """Plot a comparison between the current sensitivity indices and other ones.

        This method allows to use either a bar chart (default option) or a radar one.

        Args:
            indices: The sensitivity indices.
            output: The output
                for which to display sensitivity indices,
                either a name or a tuple of the form (name, component).
                If name, its first component is considered.
            inputs: The inputs to display. If None, display all.
            title: The title of the plot. If None, no title.
            use_bar_plot: The type of graph.
                If True, use a bar plot. Otherwise, use a radar chart.
            save: If True, save the figure.
            show: If True, show the figure.
            file_path: The path of the file to save the figures.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_format``.
            directory_path: The path of the directory to save the figures.
                If None, use the current working directory.
            file_name: The name of the file to save the figures.
                If None, use a default one generated by the post-processing.
            file_format: A file format, e.g. 'png', 'pdf', 'svg', ...
                If None, use a default file extension.
            **options: The options passed to the underlying :class:`.DatasetPlot`.

        Returns:
            A graph comparing sensitivity indices.
        """
        if not isinstance(output, tuple):
            output = (output, 0)
        if isinstance(indices, SensitivityAnalysis):
            indices = [indices]
        methods = [self] + indices
        dataset = Dataset()
        inputs_names = self._sort_and_filter_input_parameters(output, inputs)
        for name in inputs_names:
            data = abs(
                array(
                    [
                        method.main_indices[output[0]][output[1]][name]
                        for method in methods
                    ]
                )
            )
            dataset.add_variable(name, data)
        data = dataset.data[dataset.PARAMETER_GROUP]
        dataset.data[dataset.PARAMETER_GROUP] = data / data.max(axis=1)[:, None]
        dataset.row_names = [method.main_method for method in methods]
        if use_bar_plot:
            plot = BarPlot(dataset)
        else:
            plot = RadarChart(dataset)
            plot.rmin = 0.0
            plot.rmax = 1.0
        plot.title = title
        plot.execute(
            save, show, file_path, directory_path, file_name, file_format, **options
        )
        return plot

    def _save_show_plot(
        self,
        fig,  # type: Figure
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_format=None,  # type: Optional[str]
    ):  # type: (...) -> Figure
        """Save or show the plot.

        Args:
            fig: The figure to be processed.
            save: If True, save the figure.
            show: If True, show the figure.
            file_path: The path of the file to save the figures.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_format``.
            directory_path: The path of the directory to save the figures.
                If None, use the current working directory.
            file_name: The name of the file to save the figures.
                If None, use a default one generated by the post-processing.
            file_format: A file format, e.g. 'png', 'pdf', 'svg', ...
                If None, use a default file extension.

        Returns:
            The figure.
        """
        if save:
            file_path = self._file_path_manager.create_file_path(
                file_path=file_path,
                directory_path=directory_path,
                file_name=file_name,
                file_extension=file_format,
            )
        else:
            file_path = None

        save_show_figure(fig, show, file_path)
        return fig

    def export_to_dataset(self):  # type: (...) -> Dataset
        """Convert this :class:`.SensitivityIndices` instance into a :class:`.Dataset`.

        Returns:
            Dataset: The sensitivity indices.
        """
        sizes = self.dataset.sizes

        rows_names = []
        for input_name in self.inputs_names:
            for input_component in range(sizes[input_name]):
                rows_names.append("{}({})".format(input_name, input_component))

        dataset = Dataset(by_group=False)
        for method, indices in self.indices.items():
            variables = []
            sizes = {}
            data = []
            for output, components in indices.items():
                variables.append(output)
                sizes[output] = len(components)
                for component in components:
                    data.append(
                        [component[name].tolist() for name in self.inputs_names]
                    )
                    data[-1] = [item for sublist in data[-1] for item in sublist]
            data = array(data).T
            dataset.add_group(method, data, variables, sizes)
        dataset.row_names = rows_names
        return dataset



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The R2 to measure the quality of a regression algorithm.

The :mod:`~gemseo.mlearning.qual_measure.r2_measure` module
implements the concept of R2 measures for machine learning algorithms.

This concept is implemented through the :class:`.R2Measure` class
and overloads the :meth:`!MLErrorMeasure._compute_measure` method.

The R2 is defined by

.. math::

    R_2(\\hat{y}) = 1 - \\frac{\\sum_i (\\hat{y}_i - y_i)^2}
                              {\\sum_i (y_i-\\bar{y})^2},

where
:math:`\\hat{y}` are the predictions,
:math:`y` are the data points and
:math:`\\bar{y}` is the mean of :math:`y`.
"""
from __future__ import division, unicode_literals

from typing import List, NoReturn, Optional, Union

from numpy import array_split, atleast_2d
from numpy import delete as npdelete
from numpy import mean, ndarray, repeat
from sklearn.metrics import mean_squared_error, r2_score

from gemseo.mlearning.qual_measure.error_measure import MLErrorMeasure
from gemseo.mlearning.regression.regression import MLRegressionAlgo


class R2Measure(MLErrorMeasure):
    """The R2 measure for machine learning."""

    SMALLER_IS_BETTER = False

    def __init__(
        self,
        algo,  # type: MLRegressionAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for regression.
        """
        super(R2Measure, self).__init__(algo)

    def _compute_measure(
        self,
        outputs,  # type: ndarray
        predictions,  # type: ndarray
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        multioutput = "raw_values" if multioutput else "uniform_average"
        return r2_score(outputs, predictions, multioutput=multioutput)

    def evaluate_kfolds(
        self,
        n_folds=5,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        inds = samples
        folds = array_split(inds, n_folds)

        in_grp = self.algo.learning_set.INPUT_GROUP
        out_grp = self.algo.learning_set.OUTPUT_GROUP
        inputs = self.algo.learning_set.get_data_by_group(in_grp)
        outputs = self.algo.learning_set.get_data_by_group(out_grp)

        multiout = "raw_values" if multioutput else "uniform_average"

        num = 0
        ymean = mean(outputs, axis=0)
        ymean = atleast_2d(ymean)
        ymean = repeat(ymean, outputs.shape[0], axis=0)
        den = mean_squared_error(outputs, ymean, multioutput=multiout) * len(ymean)
        for n_fold in range(n_folds):
            fold = folds[n_fold]
            train = npdelete(inds, fold)
            self.algo.learn(samples=train)
            expected = outputs[fold]
            predicted = self.algo.predict(inputs[fold])
            tmp = mean_squared_error(expected, predicted, multioutput=multiout)
            num += tmp * len(fold)

        quality = 1 - num / den

        return quality

    def evaluate_bootstrap(
        self,
        n_replicates=100,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> NoReturn
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class for the estimation of various correlation coefficients."""

from __future__ import division, unicode_literals

import logging
from typing import Dict, Iterable, Mapping, Optional, Sequence, Tuple, Union

from numpy import array, vstack
from openturns import (
    CorrelationAnalysis_PCC,
    CorrelationAnalysis_PearsonCorrelation,
    CorrelationAnalysis_PRCC,
    CorrelationAnalysis_SignedSRC,
    CorrelationAnalysis_SpearmanCorrelation,
    CorrelationAnalysis_SRC,
    CorrelationAnalysis_SRRC,
    Sample,
)

from gemseo.algos.parameter_space import ParameterSpace
from gemseo.core.dataset import Dataset
from gemseo.core.discipline import MDODiscipline
from gemseo.post.dataset.radar_chart import RadarChart
from gemseo.uncertainty.sensitivity.analysis import (
    IndicesType,
    OutputsType,
    SensitivityAnalysis,
)
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.py23_compat import Path

LOGGER = logging.getLogger(__name__)


class CorrelationAnalysis(SensitivityAnalysis):
    """Sensitivity analysis based on indices using correlation measures.

    Examples:
        >>> from numpy import pi
        >>> from gemseo.api import create_discipline, create_parameter_space
        >>> from gemseo.uncertainty.sensitivity.correlation.analysis import (
        ...     CorrelationAnalysis
        ... )
        >>>
        >>> expressions = {"y": "sin(x1)+7*sin(x2)**2+0.1*x3**4*sin(x1)"}
        >>> discipline = create_discipline(
        ...     "AnalyticDiscipline", expressions_dict=expressions
        ... )
        >>>
        >>> parameter_space = create_parameter_space()
        >>> parameter_space.add_random_variable(
        ...     "x1", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x2", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x3", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>>
        >>> analysis = CorrelationAnalysis(discipline, parameter_space, n_samples=1000)
        >>> indices = analysis.compute_indices()
    """

    _PEARSON = "pearson"
    _SPEARMAN = "spearman"
    _PCC = "pcc"
    _PRCC = "prcc"
    _SRC = "src"
    _SRRC = "srrc"
    _SSRRC = "ssrrc"
    _ALGORITHMS = {
        _PEARSON: CorrelationAnalysis_PearsonCorrelation,
        _SPEARMAN: CorrelationAnalysis_SpearmanCorrelation,
        _PCC: CorrelationAnalysis_PCC,
        _PRCC: CorrelationAnalysis_PRCC,
        _SRC: CorrelationAnalysis_SRC,
        _SRRC: CorrelationAnalysis_SRRC,
        _SSRRC: CorrelationAnalysis_SignedSRC,
    }
    DEFAULT_DRIVER = "OT_MONTE_CARLO"

    def __init__(
        self,
        discipline,  # type: MDODiscipline
        parameter_space,  # type: ParameterSpace
        n_samples,  # type: int
        algo=None,  # type:Optional[str]
        algo_options=None,  # type: Optional[Mapping]
    ):  # type: (...) -> None  # noqa: D107,D205,D212,D415
        self.__correlation = None
        super(CorrelationAnalysis, self).__init__(
            discipline, parameter_space, n_samples
        )
        self.main_method = self._SPEARMAN

    @SensitivityAnalysis.main_method.setter
    def main_method(
        self,
        name,  # type: str
    ):  # type: (...) -> None
        # noqa: D102
        if name not in self._ALGORITHMS:
            methods = self._ALGORITHMS.keys()
            raise NotImplementedError(
                "{} is a bad method name. "
                "Available ones are {}.".format(name, methods)
            )
        else:
            LOGGER.info("Use {} indices as main indices.")
            self._main_method = name

    def compute_indices(
        self, outputs=None  # type: Optional[Sequence[str]]
    ):  # type: (...) -> Dict[str,IndicesType]
        # noqa: D102
        output_names = outputs or self.default_output
        if not isinstance(output_names, list):
            output_names = [output_names]
        inputs = Sample(self.dataset.get_data_by_group(self.dataset.INPUT_GROUP))
        outputs = self.dataset.get_data_by_names(output_names, True)
        self.__correlation = {}
        array_to_dict = DataConversion.array_to_dict
        for algo_name, algo_value in self._ALGORITHMS.items():
            inputs_names = self.dataset.get_names(self.dataset.INPUT_GROUP)
            sizes = self.dataset.sizes
            self.__correlation[algo_name] = {}
            for output_name, value in outputs.items():
                self.__correlation[algo_name][output_name] = []
                for index in range(value.shape[1]):
                    sub_outputs = Sample(value[:, index][:, None])
                    coefficient = array(algo_value(inputs, sub_outputs))
                    coefficient = array_to_dict(coefficient, inputs_names, sizes)
                    self.__correlation[algo_name][output_name].append(coefficient)
        return self.indices

    @property
    def pcc(self):  # type: (...) -> IndicesType
        """dict: The Partial Correlation Coefficients.

        With the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        return self.__correlation[self._PCC]

    @property
    def prcc(self):  # type: (...) -> IndicesType
        """dict: The Partial Rank Correlation Coefficients.

        With the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        return self.__correlation[self._PRCC]

    @property
    def src(self):  # type: (...) -> IndicesType
        """dict: The Standard Regression Coefficients.

        With the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        return self.__correlation[self._SRC]

    @property
    def srrc(self):  # type: (...) -> IndicesType
        """dict: The Standard Rank Regression Coefficients.

        With the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        return self.__correlation[self._SRRC]

    @property
    def ssrrc(self):  # type: (...) -> IndicesType
        """The Signed Standard Rank Regression Coefficients.

        With the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        return self.__correlation[self._SSRRC]

    @property
    def pearson(self):  # type: (...) -> IndicesType
        """dict: The Pearson coefficients.

        With the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        return self.__correlation[self._PEARSON]

    @property
    def spearman(self):  # type: (...) -> IndicesType
        """dict: The Spearman coefficients.

         ith the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        return self.__correlation[self._SPEARMAN]

    @property
    def indices(self):  # type: (...) -> Dict[str,IndicesType]
        """dict: The sensitivity indices.

        With the following structure:

        .. code-block:: python

            {
                "method_name": {
                    "output_name": [
                        {
                            "input_name": data_array,
                        }
                    ]
                }
            }
        """
        return self.__correlation

    @property
    def main_indices(self):  # type: (...) -> IndicesType # noqa: D102
        return self.__correlation[self.main_method]

    def plot(
        self,
        output,  # type: Union[str,Tuple[str,int]]
        inputs=None,  # type: Optional[Iterable[str]]
        title=None,  # type: Optional[str]
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_format=None,  # type: Optional[str]
    ):  # type: (...) -> None # noqa: D417,D102
        if not isinstance(output, tuple):
            output = (output, 0)
        dataset = Dataset()
        inputs_names = self.dataset.get_names(self.dataset.INPUT_GROUP)
        inputs_names = self._filter_names(inputs_names, inputs)
        algorithms = sorted(self._ALGORITHMS)
        data = {name: [] for name in inputs_names}
        for method in algorithms:
            indices = getattr(self, method)
            for name in inputs_names:
                data[name].append(indices[output[0]][output[1]][name])
        for name in inputs_names:
            dataset.add_variable(name, vstack(data[name]))
        dataset.row_names = algorithms
        plot = RadarChart(dataset)
        output = "{}({})".format(output[0], output[1])
        plot.title = title or "Correlation indices for the output {}".format(output)
        plot.rmin = -1.0
        plot.rmax = 1.0
        file_path = self._file_path_manager.create_file_path(
            file_path=file_path,
            directory_path=directory_path,
            file_name=file_name,
            file_extension=file_format,
        )
        plot.execute(
            save=save,
            show=show,
            file_path=file_path,
            file_name=file_name,
            file_format=file_format,
            directory_path=directory_path,
        )

    def plot_radar(
        self,
        outputs,  # type: OutputsType
        inputs=None,  # type: Optional[Iterable[str]]
        title=None,  # type: Optional[str]
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_format=None,  # type: Optional[str]
        min_radius=-1.0,  # type: float
        max_radius=1.0,  # type: float
        **options  # type:bool
    ):  # type: (...) -> RadarChart #noqa: D102
        return super(CorrelationAnalysis, self).plot_radar(
            outputs,
            inputs=inputs,
            title=title,
            save=save,
            show=show,
            file_path=file_path,
            file_name=file_name,
            file_format=file_format,
            directory_path=directory_path,
            min_radius=min_radius,
            max_radius=max_radius,
            **options
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""The Gaussian process algorithm for regression.

Overview
--------

The Gaussian process regression (GPR) surrogate model
expresses the model output as a weighted sum of kernel functions
centered on the learning input data:

.. math::

    y = \mu
        + w_1\kappa(\|x-x_1\|;\epsilon)
        + w_2\kappa(\|x-x_2\|;\epsilon)
        + ...
        + w_N\kappa(\|x-x_N\|;\epsilon)

Details
-------

The GPR model relies on the assumption
that the original model :math:`f` to replace
is an instance of a Gaussian process (GP) with mean :math:`\mu`
and covariance :math:`\sigma^2\kappa(\|x-x'\|;\epsilon)`.

Then, the GP conditioned by the learning set
:math:`(x_i,y_i)_{1\leq i \leq N}`
is entirely defined by its expectation:

.. math::

    \hat{f}(x) = \hat{\mu} + \hat{w}^T k(x)

and its covariance:

.. math::

    \hat{c}(x,x') = \hat{\sigma}^2 - k(x)^T K^{-1} k(x')

where :math:`[\hat{\mu};\hat{w}]=([1_N~K]^T[1_N~K])^{-1}[1_N~K]^TY` with
:math:`K_{ij}=\kappa(\|x_i-x_j\|;\hat{\epsilon})`,
:math:`k_i(x)=\kappa(\|x-x_i\|;\hat{\epsilon})`
and :math:`Y_i=y_i`.

The correlation length vector :math:`\epsilon`
is estimated by numerical non-linear optimization.

Surrogate model
---------------

The expectation :math:`\hat{f}` is the GPR surrogate model of :math:`f`.

Error measure
-------------

The standard deviation :math:`\hat{s}` is a local error measure
of :math:`\hat{f}`:

.. math::

    \hat{s}(x):=\sqrt{\hat{c}(x,x)}

Interpolation or regression
---------------------------

The GPR surrogate model can be regressive or interpolative
according to the value of the nugget effect :math:`\\alpha\geq 0`
which is a regularization term
applied to the correlation matrix :math:`K`.
When :math:`\alpha = 0`,
the surrogate model interpolates the learning data.

Dependence
----------
The GPR model relies on the GaussianProcessRegressor class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.gaussian_process.GaussianProcessRegressor.html>`_.
"""
from __future__ import division, unicode_literals

import logging
from typing import Callable, Iterable, Optional, Union

import openturns
from numpy import atleast_2d, ndarray
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import DataType, TransformerType
from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.utils.data_conversion import DataConversion

LOGGER = logging.getLogger(__name__)


class GaussianProcessRegression(MLRegressionAlgo):
    """Gaussian process regression."""

    LIBRARY = "scikit-learn"
    ABBR = "GPR"

    def __init__(
        self,
        data,  # type: Dataset
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        kernel=None,  # type: Optional[openturns.CovarianceModel]
        alpha=1e-10,  # type: Union[float,ndarray]
        optimizer="fmin_l_bfgs_b",  # type: Union[str,Callable]
        n_restarts_optimizer=10,  # type: int
        random_state=None,  # type: Optional[int]
    ):  # type: (...) -> None
        """
        Args:
            kernel: The kernel function. If None, use a ``Matern(2.5)``.
            alpha: The nugget effect to regularize the model.
            optimizer: The optimization algorithm to find the hyperparameters.
            n_restarts_optimizer: The number of restarts of the optimizer.
            random_state: The seed used to initialize the centers.
                If None, the random number generator is the RandomState instance
                used by `numpy.random`.
        """
        super(GaussianProcessRegression, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            kernel=kernel,
            alpha=alpha,
            optimizer=optimizer,
            n_restarts_optimizer=n_restarts_optimizer,
            random_state=random_state,
        )

        if kernel is None:
            raw_input_shape, _ = self._get_raw_shapes()
            self.kernel = Matern(
                (1.0,) * raw_input_shape, [(0.01, 100)] * raw_input_shape, nu=2.5
            )
        else:
            self.kernel = kernel

        nro = n_restarts_optimizer
        self.algo = GaussianProcessRegressor(
            normalize_y=False,
            kernel=self.kernel,
            copy_X_train=True,
            alpha=alpha,
            optimizer=optimizer,
            n_restarts_optimizer=nro,
            random_state=random_state,
        )
        self.parameters["kernel"] = self.kernel.__class__.__name__

    def _fit(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> None
        self.algo.fit(input_data, output_data)

    def _predict(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        output_pred = self.algo.predict(input_data, False)
        return output_pred

    def predict_std(
        self,
        input_data,  # type:DataType
    ):  # type: (...) -> ndarray
        """Predict the standard deviation from input data.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            output_data: The output data with shape (n_samples, n_outputs).
        """
        as_dict = isinstance(input_data, dict)
        if as_dict:
            input_data = DataConversion.dict_to_array(input_data, self.input_names)
        input_data = atleast_2d(input_data)
        inputs = self.learning_set.INPUT_GROUP
        if inputs in self.transformer:
            input_data = self.transformer[inputs].transform(input_data)
        _, output_std = self.algo.predict(input_data, True)
        return sum(output_std) / len(output_std)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Computation of tolerance intervals from a data-fitted uniform distribution."""

from gemseo.uncertainty.statistics.tolerance_interval.distribution import (
    ToleranceInterval,
)


class UniformToleranceInterval(ToleranceInterval):
    """Computation of tolerance intervals from a data-fitted uniform distribution.

    The formulae come from the R library *tolerance* [1]_.

    .. [1] Derek S. Young,
           *tolerance: An R Package for Estimating Tolerance Intervals*,
           Journal of Statistical Software, 36(5), 2010
    """

    def __init__(
        self,
        size,  # type: int
        minimum,  # type: float
        maximum,  # type: float
    ):  # type:(...) -> None
        # noqa: D205 D212 D415
        """
        Args:
            minimum: The estimation of the lower bound of the uniform distribution.
            maximum: The estimation of the upper bound of the uniform distribution.
        """
        super(UniformToleranceInterval, self).__init__(size)
        self.__minimum = minimum
        self.__maximum = maximum

    def _compute_lower_bound(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
    ):  # type: (...) -> float
        return self.__compute_exponential_bound(1 - coverage, 1 - alpha, size)

    def _compute_upper_bound(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
    ):  # type: (...) -> float
        return self.__compute_exponential_bound(coverage, alpha, size)

    def __compute_exponential_bound(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
    ):  # type: (...) -> float
        """Compute a bound of the tolerance interval for a uniform distribution.

        Args:
            coverage: A minimum percentage of belonging to the TI.
            alpha: ``1-alpha`` is the level of confidence in [0,1].
            size: The number of samples.

        Returns:
            The bound of the tolerance interval.
        """
        coefficient = coverage / alpha ** (1.0 / size)
        return (self.__maximum - self.__minimum) * coefficient + self.__minimum



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation
#               and/or initial documentation
#        :author: Sobieski, Agte, and Sandusky
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Damien Guenot
#        :author: Francois Gallard
# From NASA/TM-1998-208715
# Bi-Level Integrated System Synthesis (BLISS)
# Sobieski, Agte, and Sandusky
"""
SSBJ Aerodynamics computation
*****************************
"""
from __future__ import division, unicode_literals

import logging
from math import pi

from numpy import array, atleast_2d, cos, sin, sqrt, zeros

LOGGER = logging.getLogger(__name__)
DEG_TO_RAD = pi / 180.0


class SobieskiAerodynamics(object):
    """Class defining aerodynamical analysis for Sobieski problem and related method to
    the aerodynamical problem such as disciplines computation, constraints, reference
    optimum."""

    DTYPE_COMPLEX = "complex128"
    DTYPE_DOUBLE = "float64"

    PRESSURE_GRADIENT_LIMIT = 1.04

    def __init__(self, sobieski_base):
        """Constructor."""
        self.base = sobieski_base
        self.constants = self.base.default_constants()
        (
            self.x_initial,
            self.tc_initial,
            self.half_span_initial,
            self.aero_center_initial,
            self.cf_initial,
            self.mach_initial,
            self.h_initial,
            self.throttle_initial,
            self.lift_initial,
            self.twist_initial,
            self.esf_initial,
        ) = self.base.get_initial_values()
        self.dtype = self.base.dtype
        self.math = self.base.math

    def __set_coeff_drag_f1(self, y_32, x_2):
        """Setting of flags used for determination of polynomial coefficients.

        :param y_32: shared variables coming from blackbox_propulsion

            - y_32[0]: engine scale factor

        :type y_32: numpy array
        :param x_2: aero. design variable

            - x_2[0]: friction coeff

        :type x_2: numpy array
        :returns: s_initial, S_new, flag, bound

            - s_initial: reference value
            - s_new: new values

        :rtype: numpy array
        """
        s_initial = array([self.esf_initial, self.cf_initial], dtype=self.dtype)
        s_new = array([y_32[0], x_2[0]], dtype=self.dtype)
        flag = array([1, 1], dtype=self.dtype)
        bound = array([0.25, 0.25], dtype=self.dtype)
        return s_initial, s_new, flag, bound

    def __set_coeff_drag_f2(self, y_12):
        """Setting of flags used for determination of polynomial coefficients.

        :param y_12: shared variables coming from blackbox_structure
        :type y_12: numpy array
        :returns: s_initial, s_new, flag, bound

                - s_initial: reference value
                - s_new: new values

        :rtype: numpy array
        """
        s_initial = array([self.twist_initial], dtype=self.dtype)
        s_new = array([y_12[1]], dtype=self.dtype)
        flag = array([5], dtype=self.dtype)
        bound = array([0.25], dtype=self.dtype)
        return s_initial, s_new, flag, bound

    def __set_coeff_drag_f3(self, x_shared):
        """Setting of flags used for determination of polynomial coefficients.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: s_initial, s_new, flag, bound

                - s_initial: reference value
                - s_new: new values

        :rtype: numpy array
        """
        s_initial = array([self.tc_initial], dtype=self.dtype)
        s_new = array([x_shared[0]], dtype=self.dtype)
        flag = array([1], dtype=self.dtype)
        bound = array([0.25], dtype=self.dtype)
        return s_initial, s_new, flag, bound

    def compute_k_aero(self, x_shared):
        """Computation of a induced drag coefficient (related to lift)

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: k_aero
        :rtype: numpy array
        """
        #        if x_shared[2] >= 1:  # Mach
        #            k = (x_shared[2] ** 2 - 1) * \
        #                self.math.cos(x_shared[4] * DEG_TO_RAD) / \
        #                (4.0 * self.math.sqrt(x_shared[4] ** 2 - 1) - 2)
        #        else:
        #            k = 1. / (self.math.pi * 0.8 * x_shared[3])
        k_aero = (
            (x_shared[2] ** 2 - 1)
            * self.math.cos(x_shared[4] * DEG_TO_RAD)
            / (4.0 * self.math.sqrt(x_shared[4] ** 2 - 1) - 2)
        )
        return k_aero

    #     @staticmethod
    #     def compute_dk_aero_dAR(self, x_shared):
    #         #        if x_shared[2] >= 1:
    #         #            dk_dAR = 0.0
    #         #        else:
    #         #            dk_dAR = -1.0 / (0.8 * self.math.pi *
    # abs(x_shared[3])**2)
    #         dk_dAR = 0.0
    #         return dk_dAR

    @staticmethod
    def compute_dk_aero_dsweep(x_shared):
        """Computation of a derivative of k_aero wrt sweep.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: dk_aero_dsweep
        :rtype: numpy array
        """
        #        if x_shared[2] >= 1:
        #            dk_dsweep = -DEG_TO_RAD * (x_shared[2] ** 2 - 1) * \
        #                self.math.sin(x_shared[4] * DEG_TO_RAD) / \
        #                (4.0 * self.math.sqrt(x_shared[4] ** 2 - 1) - 2)
        #        else:
        #            dk_dsweep = 0.0
        u_velo = (x_shared[2] * x_shared[2] - 1.0) * cos(x_shared[4] * DEG_TO_RAD)
        up_velo = (
            -DEG_TO_RAD
            * (x_shared[2] * x_shared[2] - 1.0)
            * sin(x_shared[4] * DEG_TO_RAD)
        )
        v_velo = 4.0 * sqrt(x_shared[4] * x_shared[4] - 1.0) - 2.0
        vp_velo = 4.0 * x_shared[4] * (x_shared[4] * x_shared[4] - 1.0) ** -0.5
        dk_aero_dsweep = (up_velo * v_velo - u_velo * vp_velo) / v_velo ** 2
        return dk_aero_dsweep

    @staticmethod
    def compute_dk_aero_dmach(x_shared):
        """Computation of a derivative of k_aero wrt Mach.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: dk_aero_dmach
        :rtype: numpy array
        """
        # if x_shared[2] >= 1:
        #     dk_dM = abs(x_shared[3]) * (2.0 * abs(x_shared[2])) * \
        #             np.cos(x_shared[4] * np.pi / 180.) /
        # (4. * abs(x_shared[3]) * \
        #             np.sqrt(abs(x_shared[4]**2 - 1.) - 2.))
        # else:
        #     dk_dM = 0.0
        dk_aero_dmach = (
            (2.0 * x_shared[2])
            * cos(x_shared[4] * pi / 180.0)
            / (4.0 * sqrt(x_shared[4] ** 2 - 1.0) - 2.0)
        )
        return dk_aero_dmach

    def compute_dadimcf_dcf(self, x_2):
        """Computation of a derivative of adim friction coeff of polynomial wrt friction
        coeff.

        :param x_2: local design variables
        :type x_2: numpy array
        :returns: derivative of adim friction coeff of polynomial
             wrt friction coeff
        :rtype: numpy array
        """
        dadimcf_dcf = self.base.derive_normalize_s(self.cf_initial, x_2[0])
        return dadimcf_dcf

    def compute_dadimtwist_dtwist(self, y_12):
        """Computation of a derivative of adim twist of polynomial wrt twist.

        :param y_12: coupling variable from weight analysis
        :type y_12: numpy array
        :returns: derivative of adim twist of polynomial wrt twist
        :rtype: numpy array
        """
        dadimtwist_dtwist = self.base.derive_normalize_s(self.twist_initial, y_12[1])
        return dadimtwist_dtwist

    def compute_dadimesf_desf(self, y_32):
        """Computation of a derivative of adim ESF of polynomial wrt ESF.

        :param y_32: coupling variable from propulsion analysis
        :type y_32: numpy array
        :returns: derivative of adim ESF of polynomial wrt ESF
        :rtype: numpy array
        """
        return self.base.derive_normalize_s(self.esf_initial, y_32[0])

    def compute_dadimtaper_dtaper(self, x_shared):
        """Computation of a derivative of adim taper-ratio of polynomial wrt taper-
        ratio.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: derivative of adim taper-ratio of polynomial wrt taper-ratio
        :rtype: numpy array
        """
        return self.base.derive_normalize_s(self.tc_initial, x_shared[0])

    def compute_cd_min(self, x_shared, fo1):
        """Computation of a 2D minimum drag coefficient.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param fo1: coefficient for engine size
        :type fo1: numpy array
        :returns: CDmin : drag coefficient
        :rtype: numpy array
        """
        cdmin_incomp = self.constants[4]
        return (
            cdmin_incomp * fo1
            + 3.05
            * x_shared[0] ** (5.0 / 3.0)
            * (self.math.cos(x_shared[4] * DEG_TO_RAD)) ** 1.5
        )

    @staticmethod
    def compute_dcdmin_dsweep(x_shared):
        """Computation of derivative of 2D minimum drag coefficient wrt sweep.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: dCDmin_dsweep
        :rtype: numpy array
        """
        ang_rad = x_shared[4] * DEG_TO_RAD
        return (
            -3.05
            * 1.5
            * x_shared[0] ** (5.0 / 3.0)
            * cos(ang_rad) ** 0.5
            * DEG_TO_RAD
            * sin(ang_rad)
        )

    def compute_cd(self, x_shared, lift_coeff, fo1, fo2):
        """Computation of total drag coefficient.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param lift_coeff: lift coefficient
        :type lift_coeff: numpy array
        :param fo1: coefficient for engine size
        :type fo1: numpy array
        :param fo2: coefficient for twist influence on drag
        :type fo2: numpy array
        :returns: drag_coefficient : drag coefficient
        :rtype: numpy array
        """
        k_aero = self.compute_k_aero(x_shared)
        cdmin = self.compute_cd_min(x_shared, fo1)
        return fo2 * (cdmin + k_aero * lift_coeff * lift_coeff)

    def compute_drag(self, drag_coeff, x_shared):
        """Computation of drag from drag coefficient.

        :param drag_coeff: lift coefficient
        :type drag_coeff: numpy array
        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: drag as a force
        :rtype: numpy array
        """
        return self.compute_force(drag_coeff, x_shared[5], x_shared[2], x_shared[1])

    def compute_dcd_dsweep(self, x_shared, lift_coeff, fo2):
        """Computation of derivative of drag coefficient wrt sweep.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param lift_coeff: lift coefficient
        :type lift_coeff: numpy array
        :param fo2: coefficient for twist influence on drag
        :type fo2: numpy array
        :returns: dCd/dsweep
        :rtype: numpy array
        """
        dcdmin_dsweep = self.compute_dcdmin_dsweep(x_shared)
        dk_dsweep = self.compute_dk_aero_dsweep(x_shared)
        return fo2 * (dcdmin_dsweep + lift_coeff * lift_coeff * dk_dsweep)

    def compute_dcd_dsref(self, x_shared, y_12, fo2):
        """Computation of derivative of drag coefficient wrt sweep.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param y_12: lift coefficient
        :type y_12: numpy array
        :param fo2: coefficient for twist influence on drag
        :type fo2: numpy array
        :returns: dCd/dsweep
        :rtype: numpy array
        """
        k_aero = self.compute_k_aero(x_shared)
        lift_coeff = self.compute_cl(x_shared, y_12)
        dcl_dsref = self.compute_dcl_dsref(x_shared, y_12)
        return 2 * k_aero * lift_coeff * dcl_dsref * fo2

    def compute_dcd_dmach(self, x_shared, y_12, fo2):
        """Computation of derivative of drag coefficient wrt Mach.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param y_12: coupling variable from weight analysis
        :type y_12: numpy array
        :param fo2: coefficient for twist influence on drag
        :type fo2: numpy array
        :returns: dcd_dmach
        :rtype: numpy array
        """
        k_aero = self.compute_k_aero(x_shared)
        dk_dmach = self.compute_dk_aero_dmach(x_shared)
        lift_coeff = self.compute_cl(x_shared, y_12)
        dcl_dmach = self.compute_dcl_dmach(x_shared, y_12)
        return (2.0 * k_aero * dcl_dmach + lift_coeff * dk_dmach) * lift_coeff * fo2

    def compute_cl(self, x_shared, y_12):
        """Computation of lift coefficient.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param y_12: coupling variable from weight analysis
        :type y_12: numpy array
        :returns: lift coefficient Cl
        :rtype: numpy array
        """
        return self.compute_adim_coeff(y_12[0], x_shared[5], x_shared[2], x_shared[1])

    def compute_dcl_dh(self, x_shared, y_12):
        """Computation of derivative of lift coefficient wrt altitude.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param y_12: coupling variable from weight analysis
        :type y_12: numpy array
        :returns: derivative of lift coefficient wrt altitude
        :rtype: numpy array
        """
        rhov2 = self.compute_rhov2(x_shared)
        drhov2_dh = self.compute_drhov2_dh(x_shared)
        return -2 * y_12[0] / x_shared[5] * drhov2_dh / (rhov2 ** 2)

    def compute_dcl_dmach(self, x_shared, y_12):
        """Computation of derivative of lift coefficient wrt reference surface.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param y_12: coupling variable from weight analysis
        :type y_12: numpy array
        :returns: derivative of lift coefficient wrt reference surface
        :rtype: numpy array
        """
        rho, velocity = self.compute_rho_v(x_shared[2], x_shared[1])
        dv_dmach = self.compute_dv_dmach(x_shared[1])
        dcl_dmach = (
            -4.0
            * y_12[0]
            * dv_dmach
            / (rho * x_shared[5] * velocity * velocity * velocity)
        )
        return dcl_dmach

    def compute_dcl_dsref(self, x_shared, y_12):
        """Computation of derivative of lift coefficient wrt Mach number.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param y_12: coupling variable from weight analysis
        :type y_12: numpy array
        :returns: derivative of lift coefficient wrt Mach number
        :rtype: numpy array
        """
        return -self.compute_cl(x_shared, y_12) / x_shared[5]

    def compute_rhov2(self, x_shared):
        """Computation of rho * velocity * velocity (2*dynamic pressure)

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: rho * velocity * velocity
        :rtype: numpy array
        """
        rho, velocity = self.compute_rho_v(x_shared[2], x_shared[1])
        return rho * velocity * velocity

    def compute_drhov2_dh(self, x_shared):
        """Computation of derivative of rhoV**2 wrt altitude.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: drhov2_dh (derivative of rhoV**2 wrt altitude)
        :rtype: numpy array
        """
        rho, velocity = self.compute_rho_v(x_shared[2], x_shared[1])
        drho_dh, dv_dh = self.compute_drho_dh_dv_dh(x_shared[2], x_shared[1])
        return drho_dh * velocity * velocity + 2.0 * rho * dv_dh * velocity

    def compute_drhov2_dmach(self, x_shared):
        """Computation of derivative of rhoV**2 wrt Mach.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: drhov2_dmach (derivative of rhoV**2 wrt Mach)
        :rtype: numpy array
        """
        rho, velocity = self.compute_rho_v(x_shared[2], x_shared[1])
        dv_dmach = self.compute_dv_dmach(x_shared[1])
        return 2.0 * rho * dv_dmach * velocity

    def compute_rho_v(self, mach, altitude):
        """From given Mach number and altitude, compute velocity and density.

        :param mach: Mach number
        :type mach: float
        :param altitude: altitude
        :type altitude: float
        :returns: rho, velocity : density and velocity
        :rtype: float
        """
        if altitude.real < 36089.0:
            velocity = mach * 1116.39 * self.math.sqrt(1 - 6.875e-6 * altitude)
            rho = 2.377e-3 * (1 - 6.875e-6 * altitude) ** 4.2561
        else:
            velocity = mach * 968.1
            rho = 2.377e-3 * 0.2971 * self.math.exp((36089.0 - altitude) / 20806.7)
        return rho, velocity

    def compute_dv_dmach(self, altitude):
        """Computation of derivative of velocity wrt Mach.

        :param altitude: altitude
        :type altitude: numpy array
        :returns: drhov2_dmach (derivative of rhoV**2 wrt Mach)
        :rtype: numpy array
        """
        if altitude.real < 36089.0:
            dv_dmach = 1116.39 * self.math.sqrt(1 - 6.875e-6 * altitude)
        else:
            dv_dmach = 968.1
        return dv_dmach

    def compute_drho_dh_dv_dh(self, mach, altitude):
        """Compute derivative of velocity and density wrt altitude.

        :param mach: Mach number
        :type mach: float
        :param altitude: altitude
        :type altitude: float
        :returns: drho_dh, dv_dh
        :rtype: float
        """
        if altitude.real <= 36089.0:
            drho_dh = (
                -2.377e-3 * 4.2561 * 6.875e-6 * (1.0 - 6.875e-6 * altitude) ** 3.2561
            )
            dv_dh = (
                -6.875e-6 * 1116.39 * mach * 0.5 * (1.0 - 6.875e-6 * altitude) ** -0.5
            )
        else:
            drho_dh = (
                -2.377e-3
                * 0.2971
                / 20806.7
                * self.math.exp((36089.0 - altitude) / 20806.7)
            )
            dv_dh = 0.0
        return drho_dh, dv_dh

    def compute_adim_coeff(self, force, sref, mach, altitude):
        """Compute adim. force coeff from force (lift or drag)

        :param force: Force
        :type force: float
        :param sref: reference surface
        :type sref: float
        :param mach: Mach number
        :type mach: float
        :param altitude: altitude
        :type altitude: float
        :returns: force coefficient: Cl or Cd according to input force
        :rtype: float
        """
        rho, velocity = self.compute_rho_v(mach, altitude)
        return force / (0.5 * rho * velocity * velocity * sref)

    def compute_force(self, adim_coeff, sref, mach, altitude):
        """Compute force (lift or drag) from adim coeff.

        :param adim_coeff: force coefficient
        :type adim_coeff: float
        :param sref: reference surface
        :type sref: float
        :param mach: Mach number
        :type mach: float
        :param altitude: altitude
        :type altitude: float
        :returns: force: lift or drag
        :rtype: float
        """
        rho, velocity = self.compute_rho_v(mach, altitude)
        #        return adim_coeff * 0.5 * rho * V * V * sref
        return 0.5 * rho * velocity * velocity * adim_coeff * sref

    def blackbox_aerodynamics(self, x_shared, y_12, y_32, x_2, true_cstr=False):
        """This function calculates drag and lift to drag ratio of A/C.

        :param x_shared: shared design variable vector:

            - x_shared[0]: thickness/chord ratio
            - x_shared[1]: altitude
            - x_shared[2]: Mach
            - x_shared[3]: aspect ratio
            - x_shared[4]: wing sweep
            - x_shared[5]: wing surface area

        :type x_shared: numpy array
        :param y_12: shared variables coming from blackbox_structure:

            - y_12[0]: total aircraft weight
            - y_12[1]: wing twist

        :type y_12: numpy array
        :param y_32: shared variables coming from blackbox_propulsion:

            - y_32[0]: engine scale factor

        :type y_32: numpy array
        :param x_2: aero. design variable:

            - x_2[0]: friction coeff

        :type x_2: numpy array
        :param true_cstr: Default value = False)
        :returns: y_2, y_21, y_23, y_24, g_2

            - y_2: aero. analysis outputs

                - y_2[0]: lift
                - y_2[1]: drag
                - y_2[2]: lift/drag ratio

            - y_21: shared variable for blackbox_structure (lift)
            - y_23: shared variable for blackbox_propulsion (drag)
            - y_24: shared variable for BlackBoxMission (lift/drag ratio)
            - g_2: aero constraint (pressure gradient)

        :rtype: numpy array, numpy array, numpy array, numpy array, numpy array
        """

        y_2 = zeros(3, dtype=self.dtype)
        y_23 = zeros(1, dtype=self.dtype)
        y_24 = zeros(1, dtype=self.dtype)
        y_21 = zeros(1, dtype=self.dtype)
        g_2 = zeros(1, dtype=self.dtype)

        lift_coeff = self.compute_cl(x_shared, y_12)

        # Modification of CDmin for ESF and Cf
        s_initial1, s_new1, flag1, bound1 = self.__set_coeff_drag_f1(y_32, x_2)
        fo1 = self.base.poly_approx(s_initial1, s_new1, flag1, bound1)

        # Modification of drag_coeff for wing twist
        s_initial2, s_new2, flag2, bound2 = self.__set_coeff_drag_f2(y_12)
        fo2 = self.base.poly_approx(s_initial2, s_new2, flag2, bound2)

        drag_coeff = self.compute_cd(x_shared, lift_coeff, fo1, fo2)
        drag = self.compute_drag(drag_coeff, x_shared)

        y_2[1] = drag
        y_2[2] = lift_coeff / drag_coeff
        y_2[0] = y_12[0]
        y_23[0] = y_2[1]
        y_24[0] = y_2[2]
        y_21[0] = y_2[0]

        # Computation of total drag of A/C
        s_initial3, s_new3, flag3, bound3 = self.__set_coeff_drag_f3(x_shared)
        # adverse pressure gradient
        g_2[0] = self.base.poly_approx(s_initial3, s_new3, flag3, bound3)
        # Custom Pgrad function: replace by a linear function
        # coeff_dir=(1.04-0.96)/(0.06-0.04)# = 4
        #         g_2[0] = coeff_dir*Z[0]+0.8
        if not true_cstr:
            g_2[0] = g_2[0] - self.PRESSURE_GRADIENT_LIMIT
        return y_2, y_21, y_23, y_24, g_2

    @staticmethod
    def __derive_liftoverdrag(cl_cd, lift_jacobian, drag_jacobian, inv_drag):
        """Compute lift over drag jacobian terms.

        :param cl_cd: lift over drag
        :type cl_cd: numpy array
        :param lift_jacobian: d(lift)/d(var)
        :type lift_jacobian: numpy array
        :param drag_jacobian: d(drag)/d(var)
        :type drag_jacobian: numpy array
        :returns: jacobian of lift over drag
        :rtype: numpy array
        """
        return inv_drag * (lift_jacobian - cl_cd * drag_jacobian)

    @staticmethod
    def __set_coupling_jacobian(jacobian):
        """Set jacobian of coupling variables."""
        jacobian["y_21"]["x_2"] = atleast_2d(jacobian["y_2"]["x_2"][0, :])
        jacobian["y_21"]["x_shared"] = atleast_2d(jacobian["y_2"]["x_shared"][0, :])
        jacobian["y_21"]["y_12"] = atleast_2d(jacobian["y_2"]["y_12"][0, :])
        jacobian["y_21"]["y_32"] = atleast_2d(jacobian["y_2"]["y_32"][0, :])

        jacobian["y_23"]["x_2"] = atleast_2d(jacobian["y_2"]["x_2"][1, :])
        jacobian["y_23"]["x_shared"] = atleast_2d(jacobian["y_2"]["x_shared"][1, :])
        jacobian["y_23"]["y_12"] = atleast_2d(jacobian["y_2"]["y_12"][1, :])
        jacobian["y_23"]["y_32"] = atleast_2d(jacobian["y_2"]["y_32"][1, :])

        jacobian["y_24"]["x_2"] = atleast_2d(jacobian["y_2"]["x_2"][2, :])
        jacobian["y_24"]["x_shared"] = atleast_2d(jacobian["y_2"]["x_shared"][2, :])
        jacobian["y_24"]["y_12"] = atleast_2d(jacobian["y_2"]["y_12"][2, :])
        jacobian["y_24"]["y_32"] = atleast_2d(jacobian["y_2"]["y_32"][2, :])
        return jacobian

    def __initialize_jacobian(self):
        """Initialization of jacobian matrix.

        :returns:  jacobian
        :rtype: dict(dict(ndarray))
        """
        jacobian = {"y_2": {}, "g_2": {}, "y_21": {}, "y_23": {}, "y_24": {}}

        jacobian["y_2"]["x_shared"] = zeros((3, 6), dtype=self.dtype)
        jacobian["y_2"]["x_2"] = zeros((3, 1), dtype=self.dtype)
        jacobian["y_2"]["y_12"] = zeros((3, 2), dtype=self.dtype)
        jacobian["y_2"]["y_32"] = zeros((3, 1), dtype=self.dtype)

        jacobian["g_2"]["x_2"] = zeros((1, 1), dtype=self.dtype)
        jacobian["g_2"]["x_shared"] = zeros((1, 6), dtype=self.dtype)
        jacobian["g_2"]["y_12"] = zeros((1, 2), dtype=self.dtype)
        jacobian["g_2"]["y_32"] = zeros((1, 1), dtype=self.dtype)
        return jacobian

    def derive_blackbox_aerodynamics(self, x_shared, y_12, y_32, x_2):
        """This function calculates drag and lift to drag ratio of A/C.

        :param x_shared: shared design variable vector:

            - x_shared[0]: thickness/chord ratio
            - x_shared[1]: altitude
            - x_shared[2]: Mach
            - x_shared[3]: aspect ratio
            - x_shared[4]: wing sweep
            - x_shared[5]: wing surface area

        :type x_shared: numpy array
        :param y_12: shared variables coming from blackbox_structure:

            - y_12[0]: total aircraft weight
            - y_12[1]: wing twist

        :type y_12: numpy array
        :param y_32: shared variables coming from blackbox_propulsion:

            - y_32[0]: engine scale factor

        :type y_32: numpy array
        :param x_2: aero. design variable:

            - x_2[0]: friction coeff

        :type x_2: numpy array
        :returns: jacobian matrix of partial derivatives
        :rtype: dict(dict(ndarray))
        """

        jacobian = self.__initialize_jacobian()
        lift_coeff = self.compute_cl(x_shared, y_12)
        # Modification of drag_coeff_min for ESF and Cf
        s_initial1, s_new1, flag1, bound1 = self.__set_coeff_drag_f1(y_32, x_2)
        fo1, ai_coeff1, aij_coeff1, s_shifted1 = self.base.derive_poly_approx(
            s_initial1, s_new1, flag1, bound1
        )

        drag_coeff_incomp = self.constants[4]
        drag_coeff_min = self.compute_cd_min(x_shared, fo1)

        k_aero = self.compute_k_aero(x_shared)

        # Modification of drag_coeff for wing twist
        s_initial2, s_new2, flag2, bound2 = self.__set_coeff_drag_f2(y_12)
        fo2, ai_coeff2, aij_coeff2, s_shifted2 = self.base.derive_poly_approx(
            s_initial2, s_new2, flag2, bound2
        )

        drag_coeff = self.compute_cd(x_shared, lift_coeff, fo1, fo2)
        cl_cd = lift_coeff / drag_coeff
        cl2 = lift_coeff * lift_coeff
        drag = self.compute_drag(drag_coeff, x_shared)

        rhov2 = self.compute_rhov2(x_shared)
        dyn_pressure = 0.5 * rhov2
        dyn_force = dyn_pressure * x_shared[5]
        inv_drag = 1.0 / drag

        # dLift_dCf
        #        jacobian['y_2']['x_2'][0, 0] = 0.0

        # dDrag_dCf
        dadimcf_dcf = self.compute_dadimcf_dcf(x_2)

        dy2dx2 = dyn_force * fo2 * drag_coeff_incomp * dadimcf_dcf
        dy2dx2 *= (
            ai_coeff1[1]
            + aij_coeff1[1, 0] * s_shifted1[0]
            + aij_coeff1[1, 1] * s_shifted1[1]
        )  # dDrag_dCf
        jacobian["y_2"]["x_2"][1, 0] = dy2dx2
        # d(Lift/Drag)_dCf

        dlod = self.__derive_liftoverdrag(
            cl_cd, jacobian["y_2"]["x_2"][0, 0], jacobian["y_2"]["x_2"][1, 0], inv_drag
        )
        jacobian["y_2"]["x_2"][2, 0] = dlod

        # dLift/dZ= 0.0
        # dDrag/d(t/c)
        dy2dxs = dyn_force * fo2 * 3.05 * 5.0 / 3.0 * x_shared[0] ** (2.0 / 3.0)
        dy2dxs *= (self.math.cos(x_shared[4] * DEG_TO_RAD)) ** 1.5
        jacobian["y_2"]["x_shared"][1, 0] = dy2dxs

        # d(Lift/Drag)/d(t/c)

        # d(Drag)/dh
        drhov2_dh = self.compute_drhov2_dh(x_shared)
        dcl_dh = self.compute_dcl_dh(x_shared, y_12)
        dy2dxs = 0.5 * x_shared[5] * fo2 * drag_coeff_min * drhov2_dh
        add_der = cl2 * drhov2_dh + rhov2 * 2.0 * lift_coeff * dcl_dh
        dy2dxs += 0.5 * x_shared[5] * k_aero * fo2 * add_der
        jacobian["y_2"]["x_shared"][1, 1] = dy2dxs
        # d(Lift/Drag)/dh

        # d(Drag)/dM
        dcd_dmach = self.compute_dcd_dmach(x_shared, y_12, fo2)
        drhov2_dmach = self.compute_drhov2_dmach(x_shared)
        jacobian["y_2"]["x_shared"][1, 2] = (
            0.5 * x_shared[5] * (drag_coeff * drhov2_dmach + rhov2 * dcd_dmach)
        )
        # d(Lift/Drag)/dM

        # d(Drag)/dAR
        #         dk_dAR = self.compute_dk_aero_dAR(x_shared) # = 0.0
        #         jacobian['y_2']['x_shared'][1, 3] = dyn_force * dk_dAR * fo1 * cl2
        # d(Lift/Drag)/dAR

        # d(Drag)/dsweep
        dcd_dsweep = self.compute_dcd_dsweep(x_shared, lift_coeff, fo2)
        jacobian["y_2"]["x_shared"][1, 4] = dyn_force * dcd_dsweep

        # d(Drag)/dsref
        dcd_dsef = self.compute_dcd_dsref(x_shared, y_12, fo2)
        dy2dxs = drag / x_shared[5] + dyn_force * dcd_dsef
        jacobian["y_2"]["x_shared"][1, 5] = dy2dxs

        for i in range(6):
            dy2dxs0i = jacobian["y_2"]["x_shared"][0, i]
            dy2dxs1i = jacobian["y_2"]["x_shared"][1, i]
            dlod = self.__derive_liftoverdrag(cl_cd, dy2dxs0i, dy2dxs1i, inv_drag)
            jacobian["y_2"]["x_shared"][2, i] = dlod

        # dLift/dWt
        jacobian["y_2"]["y_12"][0, 0] = 1.0
        # d(Drag)/dWt
        jacobian["y_2"]["y_12"][1, 0] = 2.0 * y_12[0] * k_aero * fo2 / dyn_force
        # d(Lift/Drag)/dWt
        #         jacobian['y_2']['y_12'][
        #             2, 0] = self.__derive_liftoverdrag(
        #             cl_cd, jacobian['y_2']['y_12'][0, 0],
        #             jacobian['y_2']['x_shared'][1, 0], inv_drag)

        # dLift/dtwist= 0.0
        # d(Drag)/dtwist
        dadimtwist_dadim = self.compute_dadimtwist_dtwist(y_12)
        jacobian["y_2"]["y_12"][1, 1] = (
            dyn_force
            * dadimtwist_dadim
            * (drag_coeff_min + k_aero * lift_coeff * lift_coeff)
            * (ai_coeff2[0] + aij_coeff2[0, 0] * s_shifted2[0])
        )
        # d(Lift/Drag)/dtwist
        #         jacobian['y_2']['y_12'][
        #             2, 1] = inv_drag * (jacobian['y_2']['y_12'][0, 1] -
        #                                 cl_cd * jacobian['y_2']['y_12'][1, 1])
        for i in range(2):
            dy2dy120 = jacobian["y_2"]["y_12"][0, i]
            dy2dy121 = jacobian["y_2"]["y_12"][1, i]
            dy2dy122 = self.__derive_liftoverdrag(cl_cd, dy2dy120, dy2dy121, inv_drag)
            jacobian["y_2"]["y_12"][2, i] = dy2dy122

        # dLift/dESF
        #        jacobian['y_2']['y_32'][0, 0] = 0.0
        # dDrag/dESF
        dadimesf_desf = self.compute_dadimesf_desf(y_32)
        jacobian["y_2"]["y_32"][1, 0] = (
            dyn_force
            * fo2
            * drag_coeff_incomp
            * dadimesf_desf
            * (
                ai_coeff1[0]
                + aij_coeff1[0, 0] * s_shifted1[0]
                + aij_coeff1[0, 1] * s_shifted1[1]
            )
        )
        # d(Lift/Drag)/dtwist
        jacobian["y_2"]["y_32"][2, 0] = self.__derive_liftoverdrag(
            cl_cd,
            jacobian["y_2"]["y_32"][0, 0],
            jacobian["y_2"]["y_32"][1, 0],
            inv_drag,
        )

        # d(dp/dx)/d(t/c)
        s_initial3, s_new3, flag3, bound3 = self.__set_coeff_drag_f3(x_shared)
        _, ai_coeff3, aij_coeff3, s_shifted3 = self.base.derive_poly_approx(
            s_initial3, s_new3, flag3, bound3
        )

        dadimtaper_dtaper = self.compute_dadimtaper_dtaper(x_shared)
        jacobian["g_2"]["x_shared"][0, 0] = dadimtaper_dtaper * (
            ai_coeff3[0] + aij_coeff3[0, 0] * s_shifted3[0]
        )
        #        jacobian['g_2']['x_shared'][0, 1:] = 0.0
        #        jacobian['g_2']['x_2'][0, :] = 0.0
        #        jacobian['g_2']['y_12'][0, :] = 0.0
        #        jacobian['g_2']['y_32'][0, :] = 0.0

        jacobian = self.__set_coupling_jacobian(jacobian)

        return jacobian
